// PlayerSkills.cpp

// Implementovanie schopnosti hraca, ktore sa potom vyuziju pri taktizovani.

#include "Player/PlayerSkills.h"
//#include "Common/Main.h"
#include "Common/Tasks.h"
//#include "Common/Support/Debug.h"
//#include <stdio.h>
#include <math.h>
//#include <assert.h>
//#include <string.h>
//#include <limits.h>

#include <string>
#include <fstream>

using namespace std;

// Premenna signalizujuca funkcii NeckSynchronizationWithVI, ze prisla nova VI a treba okamzite poslat prikaz TurnNeck na server
#define NEW_VI 1000
#define NAR_VIEW 1001
#define NOR_VIEW 1002
#define ALIG_NECK 1003

// Po kolkych cyklov od okamihu chytenia lopty pokracujeme v simulacii?
#define AFTER_CATCH_CYCLES 5
//
#define BALL_POS_TOL 5.0f
//
#define DRIBBLE_STOP_DIST 10.0f
// Kompenzia za chybu vnimania pozicie lopty
#define POS_ERROR_DELTA 3

#define max(x,y) ( ((x) > (y)) ? (x) : (y) )
#define min(x,y) ( ((x) > (y)) ? (y) : (x) )
#define sqr(x) (x)*(x)

// Globalne pomocne metody

// Zoznam informacii o hracoch, ktorych vidi hrac
extern CEvaluatingPlayerList* ePlayerList;

// Vrati uhol vstupneho vektora zmenseni o uhol zadany na vstupe.
static Angle CalcDir(const Vector& vect, Angle rotation)
{
	Vector tmp = vect - rotation;
	return tmp.GetAngle();
}

// Prelinaju sa ciary?
static inline bool DoLinesIntersect(Point A, Vector U, Point B, Vector V)
{
	// Do lines (A + t1 * U) and (B + t2 * V) intersect?
	float t1 = (U.GetX() * B.GetY() - U.GetX() * A.GetY() - U.GetY() * B.GetX() + U.GetY() * A.GetX()) / (U.GetY() * V.GetX() - U.GetX() * V.GetY());
	float t2 = (V.GetX() * A.GetY() - V.GetX() * B.GetY() - V.GetY() * A.GetX() + V.GetY() * B.GetX()) / (V.GetY() * U.GetX() - V.GetX() * U.GetY());

	return t1 >= 0.f && t1 <= 1.f && t2 >= 0.f && t2 <= 1.f;
}

// Prevod desatineho cisla na cele
// Priklad zaokruhleneho vstupu: (0.49 -> 0, 0.5 -> 1)
inline static int f2int(float fX)
{ 
	return int(fX + 0.5f);
}

// Vrati, ci ma alebo nema priamka priesecnik s kruznicou. Vypocitany priesecnik je ulozeny do T1.
// POZN.: Tato funkcia je divna. Priamka moze mat aj 2 priesecniky s kruznicou.
int GetLineCircleIntersect(Point A, Point B, Point S, float r, Point &T1)
{
	static float Ax, Ay, Bx, By, Sx, Sy;
	static float a, b, c, t;
	Ax = A.GetX(); Bx = B.GetX(); Sx = S.GetX();
	Ay = A.GetY(); By = B.GetY(); Sy = S.GetY();
	a = 2 * (sqr(Ax-Bx) + sqr(Ay-By));
	
	if(fabsf(a) < .0001f)
		return 0;

	b = -4*(sqr(Ax-Bx) + sqr(Ay-By))*(sqr(Ax)-sqr(r)-2*Ax*Sx+sqr(Sx)+sqr(Ay-Sy));
	b += 4*sqr((sqr(Ax)+sqr(Ay)+Bx*Sx-Ax*(Bx+Sx)+By*Sy-Ay*(By+Sy)));
	
	if(b < 0)
		return 0;
	b = sqrtf(b);

	c = 2*(Ax-Bx)*(Ax-Sx)+2*(Ay-By)*(Ay-Sy);

	t = (c - b) / a;
	if(t > 0)
	{
		T1.SetX(Ax + (Bx-Ax)*t);
		T1.SetY(Ay + (By-Ay)*t);
		return 1;
	}
	t = (c + b) / a;
	if(t > 0)
	{
		T1.SetX(Ax + (Bx-Ax)*t);
		T1.SetY(Ay + (By-Ay)*t);
		return 1;
	}
	return 0;
}

// Vyriesi kvadraticku rovnicu
// a,b,c su koeficienty rovnice, x1, x2 su vysledky kvadratickej rovncie
// vrati true, ak x1 aj x2 su platne, inak false (determinant < 0)
inline static bool Quadrat(float a, float b, float c, float* x1, float* x2)
{
	const float d = b*b - 4*a*c;
	
	// diskriminant zaporny
	if(d < 0)
		return false;

	const float sq_d = sqrtf(d);

	*x1 = (-b + sq_d) / (2*a);
	*x2 = (-b - sq_d) / (2*a);
	
	return true;
}

// -------------------------------------- PlayerSkills --------------------------------------

// Konstruktor

PlayerSkills::PlayerSkills(const char* name, const char* desc) : PlayerKernel("PlayerKernel", "PlayerKernel module")
{
	m_pCommandGen = NULL;
	m_bKickable = false;
	m_PlayerRole = PR_Player;
	m_iGoalieCatchBan = 0;

	m_pFCDribbleNet = NULL;
	FCWriteTime = 0;
	FCDiffTime = 0;

	TurnNeckTo_futurePos.SetX(0), TurnNeckTo_futurePos.SetY(0);
	TurnNeckTo_futureMyPos.SetX(0), TurnNeckTo_futureMyPos.SetY(0);
	TurnNeckTo_nextTurn = 0;
	nNeckDirection = -1;
	bAvoidNextCorrection = false;
	nChangeView = 0;

// --- FC Farmari ---
	
	// Vytvorenie siete aj s nastavenim poctu jej neuronov...	
	m_pFCDribbleNet = new FCNetwork(DRIBBLE_NET_INPUT, 30, 0.5, 0.85);
	m_pFCDribbleNet->Inicialize();
	m_pFCDribbleNet->LoadFrom("vahydrib.txt");
}

// Destruktor
PlayerSkills::~PlayerSkills()
{
	// --- FC Farmari ---

	if (m_pFCDribbleNet)
	{
		delete m_pFCDribbleNet;
		m_pFCDribbleNet = NULL;
	}
	
}

// Inicializacia rozhrania
int PlayerSkills::Init(int argc, char** argv, int iTask)
{
	m_pCommandGen = (CommandsInterface*) Main::GetInstance()->GetInterface(COMMANDS_TASK);
	
	if (m_pCommandGen == NULL)
		return INIT_ERR;

	return INIT_OK;
}


// PRIVATE METODY

// Pomocna metoda pre driblovanie
// POZN.: Cela funkcia je neprehladna a s velkou pravdepodobnostou nepresna, odporucam nepouzivat!
bool PlayerSkills::DribbleRun(float angle)
{
	SKILLS_DBGN("%i: ->dribble run", world->state->GetSimTime());

	const float karea = world->params->server->ball_size + world->me->type->player_size + world->me->type->kickable_margin;
	
	// POZN.: Nepresny (resp. hruby) vypocet dalsej pozicie lopty. (chyba vo vzorci ball_decoy)
	const Point ball_nextpos = world->ball->pos + world->ball->vel;

	// Loptu nema sancu dostihnut za ziadnu cenu
	if((ball_nextpos - world->me->pos).GetLength() > (world->me->type->player_speed_max + karea))
		return false;
	
	const Point A = world->me->pos + world->me->vel;
	const float dashpower = RecommDashPower(true);
	const float maxacc = dashpower * world->me->type->dash_power_rate * world->me->GetEffort();
	float u = world->me->GetAbsDir().GetRad();

	const Point B = A + RectVector(maxacc * cosf(u), maxacc * sinf(u));

	// Vzdialenost bodu od priamky
	float sDS =  ((B.GetX() - A.GetX())*(ball_nextpos.GetY() - A.GetY()) - (B.GetY() - A.GetY())*(ball_nextpos.GetX() - A.GetX())) / maxacc;
	float DS = fabsf(sDS);

	if(DS > (0.9f * karea))
	{
		// Na loptu nedociahne, resp. riziko, ze utecie
		return false;
	}

	const float controll_dist = world->params->server->ball_size + world->me->type->player_size + 0.45f * world->me->type->kickable_margin;
	
	if(fabsf(sDS - controll_dist*sinf(angle*M_PI/180)) > 0.4f)
	{
		return false;
	}

	float T = ( (B.GetX() - A.GetX())*(ball_nextpos.GetX() - A.GetX()) + (B.GetY() - A.GetY())*(ball_nextpos.GetY() - A.GetY()) ) / (maxacc*maxacc);

	if(T > 1)
	{
		if( (ball_nextpos - B).GetLength() > karea)
		{
			// Lopta je prilis daleko
			return false;
		}
	}
	else if(T <= 0)
	{
		// Lopta je prilis daleko alebo sa absolutne neoplati pohnut
		return false;
	}

	// Ak je prilis blizko
	if(DS < controll_dist )
	{
		float dt = sqrtf( controll_dist*controll_dist - DS*DS )/ maxacc;
		T -= dt;
	}

	if(T > 1)
	{
		T = 1;
	}
	else if(T <= 0)
	{
		return false;
	}

	if(T > 0.5f)
	{
		float power = T * dashpower;
		m_pCommandGen->Dash(power);
		TurnNeckTo(ball_nextpos, world->me->pos + PredictNextVel(power),0);
		return true;
	}
	
	// Bezalo by sa prilis pomaly, neoplati sa bezat
	else 
	{
		return false;
	}
}

// Kopne loptu na urcite miesto
// POZN.: Lopta v cielovom bode zastane alebo prejde cez ten bod nejakou rychlostou?
// Najlepsie tu funkciu doplnit o parameter rychlost lopty cez bod a funkciu prepracovat.
void PlayerSkills::NearKick(const Point& ciel)
{
	SKILLS_DBGN("%i: ->near kicking to position", world->state->GetSimTime());
	SKILLS_DBGN("%i: ->pos (%3.2f, %3.2f)", world->state->GetSimTime(), ciel.GetX(), ciel.GetY());

	// Ak je lopta dostatocne blizko
	if(m_fBallDistance < (world->me->type->player_size + world->me->type->kickable_margin + world->params->server->ball_size))
	{
		if(m_fBallDistance < (world->me->type->player_size + world->params->server->ball_size))
		{
			// Nemoze byt tak blizko
			m_fBallDistance = world->me->type->player_size + world->params->server->ball_size;
		}

		float dir_diff = (m_BallDiff.GetAngle() - world->me->GetAbsDir()).GetAbsDeg();
		const float ball_power_rate = world->params->server->kick_power_rate * (1 - 0.25f * (dir_diff/180) - 0.25f * (m_fBallDistance - world->me->type->player_size - world->params->server->ball_size) / world->me->type->kickable_margin);

		float maxacc = world->params->server->maxpower * ball_power_rate;
		if(maxacc > world->params->server->ball_accel_max)
		{
			maxacc = world->params->server->ball_accel_max;
		}

		Point ball_zac = (world->ball->pos + world->ball->vel);
		Vector k = ciel - ball_zac;

		float req_acc = 1.14f * k.GetLength();
		float kick_dir = CalcDir(k, world->me->GetAbsDir()).GetDeg(); // povodny vypocet smeru

		if(req_acc > maxacc) // tak silno nedokaze kopnut
		{

			Vector smer = ciel - world->ball->pos;
			const float vzdial = smer.GetLength();

			float a = vzdial * vzdial;
			float b = -2 * world->ball->vel.GetX() * smer.GetX()
				-2 * world->ball->vel.GetY() * smer.GetY();
			float c = world->ball->vel.GetX() * world->ball->vel.GetX() + world->ball->vel.GetY() * world->ball->vel.GetY() - maxacc*maxacc;

			float n1,n2;
			bool success = Quadrat(a,b,c,&n1,&n2);
			if(success) // kvadraticka rovnica ma riesenie
			{
				//riesenie je pouzitelne
				if(n1 >= 0 && n1 <= 1)
				{
					Vector nkdir = smer * n1 - world->ball->vel;
					kick_dir = CalcDir(nkdir, world->me->GetAbsDir()).GetDeg();
				}
				// inak sa pouzije prednastavena hodnota kick_dir
			}
			else //pokus o spomalenie lopty
			{
				kick_dir = CalcDir(world->ball->vel * -1.0f, world->me->GetAbsDir()).GetDeg();
			}

			req_acc = maxacc;
		}

		float req_power = req_acc / ball_power_rate;
		m_pCommandGen->Kick(req_power,kick_dir);
	}
}

// Rozhodovanie sa mimo modu play_on (cas je zastaveny).
// POZN.: Vyuzit zastavenie casu na pozorovanie okolia! (napr. 1x sa otocit okolo vlastnej osi)
void PlayerSkills::RunPlayOff()
{
	PlayMode pmode = world->state->GetPlayMode();
	if(pmode == PM_BeforeKickOff || pmode == PM_AfterGoal_Our || pmode == PM_AfterGoal_Their)
	{
		Point home = GetHomePos();

		if((home - world->me->pos).GetLength() > POS_TOL)
		{
			MoveTo(home);
		}
		else
		{
			TurnTo(Point(0, 0));
		}

		NeckSynchronizationWithVI(ALIG_NECK);

    }
	else
	{
		// Pravdepodobne som v ofsajde
#ifndef GOALIE
		
		const float off_l = GetOffsideLine();
		
		if(world->me->pos.GetX() > off_l - 0.1f)
		{
			world->me->pos.SetX(off_l - 0.1f);
		}

		// Odbehni z offsajdu
		RunToPosition(world->me->pos, 1);
		NeckSynchronizationWithVI(ALIG_NECK);
#endif
	}
}

// Rozhodovanie sa pocas modu play_on.	
// POZN.: Ma volanie SB zmysel? Dostane hrac SB viackrat za cyklus?
void PlayerSkills::RunPlayOn()
{
//	if(GetPlayerRole() == PR_Goalie)
//		DBGN("%d: BallAge: %d, valid: %s", world->state->GetSimTime(), world->ball->GetAge(), world->ball->IsValid() ? "true" : "false");
	if(world->ball->IsValid() || (GetPlayerRole() == PR_Goalie && world->ball->GetAge() < 11))
	{
		switch(GetPlayerRole())
		{
		// Brankar
		case PR_Goalie: 
			{
				if(CanCatchBall() || IsBallKickAble() || CanTackleBall() || world->state->GetPlayMode() == PM_GoalKick_Our)
				{
					BehaveWithBallGoalie();
				}
				else
				{
					BehaveWithoutBallGoalie();
				}
			}
			break;

		// Regularny hrac
		default:
			{
				if(IsBallKickAble())
				{
					BehaveWithBall();
				}
				else
				{
					BehaveWithoutBall();
				}
			}
			break;
		}

		// Poslem spravu svojim spoluhracom
		SendInfo();

		if((!world->ball->IsValid() || world->ball->GetAge() > 3) && (world->ball->pos - world->me->pos).GetLength() > 4)
		{
			m_pCommandGen->ChangeView(WIDE_VIEW);
			SetNextSynchMode(Synch_Normal);
		}
		else
		{
			ViewWidth odporucana_sirka = WidthFromDistance(m_fBallDistance);
			if(m_PlayerRole == PR_Goalie && GetNextSynchMode() == Synch_WaitForVI)
			{
				m_pCommandGen->ChangeView(WIDE_VIEW);
			}
			else if(odporucana_sirka != world->me->GetViewWidth())
			{
				switch(odporucana_sirka)
				{
				case WIDE_VIEW:
					m_pCommandGen->ChangeView(WIDE_VIEW);
					SetNextSynchMode(Synch_Normal);
					break;
				case NORMAL_VIEW:
					NeckSynchronizationWithVI(NOR_VIEW);
					SetNextSynchMode(Synch_Normal);
					break;
				case NARROW_VIEW:
					NeckSynchronizationWithVI(NAR_VIEW);
					SetNextSynchMode(Synch_VIOnly);
					break;
				}	
			}
		}
	}
	// Nemame informaciu o lopte
	else 
	{
#ifdef GOALIE
		BehaveWithoutBallKnowledgeGoalie();
#else // GOALIE
		// Hlada loptu
		Scan();
#endif //GOALIE
	}
}

// PROTECTED METODY

// BEHACIE (PRESUVACIE) METODY
// ---------------------------

// Presun sa na urcity bod (skokova metoda)
void PlayerSkills::MoveTo(const Point& pos)
{
	m_pCommandGen->Move(pos.GetX(), pos.GetY());
}

// Bezi na zadanu poziciu, na ktorej ocakava najvhodnejsie zachytenie lopty 
void PlayerSkills::RunToBallInteceptPosition(Point position)
{
	int nsteps = -1;
	int psteps;
	Angle uhol;
	Angle uholT = DegAngle(0);
	Ball tmpBall(*world->ball);
	tmpBall.SetPosition(position);
	Vector diff;

	float dashpower;
	
	// Bezi sa co narychlejsie, ak zachytavame prihravku hlboko v obrane supera alebo nasej obrane
	if(IsPointInRectArea(world->ball->pos.GetX(), world->ball->pos.GetY(), 28.0, -34.0, 52.5, 34.0) ||
		IsPointInRectArea(position.GetX(), position.GetY(), -52.5, -34.0, -25.0, 34.0))
	{
		
#ifdef GOALIE
		if(IsPointInRectArea(position.GetX(), position.GetY(), -52.5, -20.0, -27, 20.0))
		{	
			DBGN("\n[Cyklus %d] Pouzivam zrychlene zachytavanie lopty na pozicii: (%lf, %lf)", world->state->GetSimTime(), world->ball->pos.GetX(), world->ball->pos.GetY());
			dashpower = RecommendDashToMaxSpeed();
		}
		else
			dashpower = RecommDashPower(true);
#else	// GOALIE
		if(ShouldInterceptBall())
			dashpower = RecommendDashToMaxSpeed();
		else
			dashpower = RecommDashPower(false);
#endif	// GOALIE
	
	}
	else
		dashpower = RecommDashPower(true);

	if((world->ball->pos - world->me->pos).GetLength() < 6.0)
		dashpower = RecommendDashToMaxSpeed();

	do
	{
		diff = tmpBall.pos - world->me->pos;
		uhol = CalcDir(diff, world->me->GetAbsDir());
		
		if(uhol.GetAbsDeg() > 10)
		{
			diff = tmpBall.pos - (world->me->pos + PredictNextVel(dashpower));
			uholT = CalcDir(diff, world->me->GetAbsDir());
			psteps = 1 + PredictSteps(diff.GetLength(), dashpower);
		}
		else
		{
			psteps = PredictSteps(diff.GetLength(), dashpower, PredictNextVel(dashpower).GetLength());
		}
		
		nsteps++;
		tmpBall.Step();
	}
	while(nsteps < psteps);

	Ball tmpBall2(*world->ball);	
	tmpBall2.Step();

	if(uhol.GetAbsDeg() > 10)
	{
		float TurnMoment = CompMoment(uholT.GetDeg());
		m_pCommandGen->Turn(TurnMoment);
		TurnNeckTo(tmpBall2.pos, world->me->pos + world->me->vel, CompTurn(TurnMoment));
	}
	else
	{
		m_pCommandGen->Dash(dashpower);
		TurnNeckTo(tmpBall2.pos, world->me->pos + (PredictNextVel(dashpower) * 1.1f /* Bulharska konstanta */) , 0);
	}	
}


// Bezi sa do bodu s istym natocenim.
void PlayerSkills::RunToPoint(const Point& target, bool hurry, bool forLife, float posTol, int angTol)
{
	
	float power;
			
	if(forLife)
		power = RecommendDashToMaxSpeed();
	else 
		power = RecommDashPower(hurry);

	// Kde zastaneme ?
	Point stopPos = world->me->pos + PredictNextVel(power);
	Vector diff = target - stopPos;

	Ball tmpBall(*world->ball);
	tmpBall.Step();

	if(diff.GetLength() > posTol)
	{
		// Sme prilis daleko
//		float angle = CalcDir(target - world->me->pos, world->me->GetAbsDir());

//		Point futurePos = world->me->pos + PredictNextVel(power);
//		float turnMoment = CompMoment(CalcDir(target - futurePos, world->me->GetAbsDir()).GetDeg());

		float turnMoment = CompMoment(CalcDir(target - world->me->pos, world->me->GetAbsDir()).GetDeg());

		if(fabsf(turnMoment) > angTol)
		{
			// Musime sa otocit
			m_pCommandGen->Turn(turnMoment);

			TurnNeckTo(tmpBall.pos, world->me->pos + world->me->vel, CompTurn(turnMoment));
		}
		else
		{
			// Nemusime sa otocit
			m_pCommandGen->Dash(power);
			TurnNeckTo(tmpBall.pos, world->me->pos + PredictNextVel(power), 0);
		}
	}
}

// Bez na zadnu poziciu celom vpred (pre brankara)
void PlayerSkills::RunToPointBackwards(const Point& target, bool hurry, bool forLife, float posTol, int angTol)
{
	float power;
			
	if(forLife)
		power = RecommendDashToMaxSpeed();
	else 
		power = RecommDashPower(hurry);

	// Kde zastaneme?
	Point stopPos = world->me->pos + PredictNextVel(power);
	Vector diff = target - stopPos;

	Ball tmpBall(*world->ball);
	tmpBall.Step();

	if(diff.GetLength() > posTol)
	{
		// Sme prilis daleko
		float angle = CalcDir(target - world->me->pos, world->me->GetAbsDir()).GetDeg();

		if(fabsf(angle) < 180 - angTol)
		{
			// Musime sa otocit
//			Point futurePos = world->me->pos + PredictNextVel(power);
//			float turnMoment = CompMoment(CalcDir(RectVector(0, 0) - (target - futurePos), world->me->GetAbsDir()).GetDeg());
			
			float turnMoment = CompMoment(CalcDir(RectVector(0, 0) - (target - world->me->pos), world->me->GetAbsDir()).GetDeg());
			
			m_pCommandGen->Turn(turnMoment);

			TurnNeckTo(tmpBall.pos, world->me->pos + world->me->vel, CompTurn(turnMoment));
		}
		else
		{
			// Nemusime sa otocit
			m_pCommandGen->Dash(-power * .5f);
			TurnNeckTo(tmpBall.pos, world->me->pos + PredictNextVel(power), 0);
		}
	}
}

// Bezi na zadanu poziciu, na ktorej ocakava najvhodnejsie zachytenie lopty 
void PlayerSkills::RunToPosition(Point position, int nSpeedMode)
{
	float dashpower;

	switch(nSpeedMode)
	{
	case 0:
		dashpower = RecommDashPower(false);
		break;
	case 1:
		dashpower = RecommDashPower(true);
		break;
	case 2:
		dashpower = RecommDashPower(true, true);
		break;
	default:
		dashpower = RecommDashPower(true);
		break;
	}

	// Vypocet ofsajdu
	float offside_line = 0;
	float fast = 0;
	float second = 0;
	float tmp;
	int i;

	for (i = 0; i < world->enemies.GetSize(); i++)
	{	// Druhy obranca
		if (world->enemies[i].pos.GetX() > second)
		{
			second = world->enemies[i].pos.GetX();
			
			if (second > fast)
			{
				tmp = fast;
				fast = second;
				second = tmp;
			}
		}
	}

	if (second > world->ball->pos.GetX())
	{
		offside_line = second;
	}
	else
	{
		offside_line = world->ball->pos.GetX();
	}

	
	Point tmpPosition = position;
	
	if(tmpPosition.GetX() > offside_line)
		tmpPosition.SetX(offside_line - 1.0f);

	Vector diff = tmpPosition - world->me->pos;
	Angle uhol = CalcDir(diff, world->me->GetAbsDir());
	
	Ball tmpBall(*world->ball);	
	tmpBall.Step();

	if(diff.GetLength() > 0)
	{
		if(uhol.GetAbsDeg() > 10)
		{
			float turnMoment = CompMoment(uhol.GetDeg());
			m_pCommandGen->Turn(turnMoment);
			TurnNeckTo(tmpBall.pos, world->me->pos + world->me->vel, CompTurn(turnMoment));
		}
		else
		{
			m_pCommandGen->Dash(dashpower);
			TurnNeckTo(tmpBall.pos, world->me->pos +  PredictNextVel(dashpower), 0);
		}
	}
}

// CHYTACIE METODY
// ---------------

// Dokazeme chytit loptu?
// POZN.: Tie konstanty pri chytani lopty su nejake divne (asi rataju aj s chybou pozicie)
bool PlayerSkills::CanCatchBall() const
{
	// Iba brankar moze chytat loptu
	if(m_PlayerRole != PR_Goalie)
		return false;

	// V tomto mode sa nemoze chytat
	if(world->state->GetPlayMode() != PM_PlayOn)
		return false;

	// Lopta je daleko
	if((world->ball->pos - world->me->pos).GetLength() >= 2.0f)
		return false;

	// Pred chvilou chytal
	if(m_iGoalieCatchBan)
		return false;

	// Je mimo penaltoveho uzemia
	if(fabsf(world->me->pos.GetY()) >= 20.16f)
		return false;
	if(world->ball->pos.GetX() >= -36.f)
		return false;

	return true;
}

// Chyti loptu
void PlayerSkills::CatchBall()
{
	float uhol = CalcDir(world->ball->pos - world->me->pos, world->me->GetAbsDir()).GetDeg();
	m_pCommandGen->Catch(uhol);
	m_iGoalieCatchBan = world->params->server->catch_ban_cycle; 
}


// DRIBLOVACIE METODY
// ------------------
// 1. sposob driblovania
// POZN.: Zbytocna funkcia nepouzivat. Neriesi interval (0 < vzdial <= BALL_POS_TOL)
void PlayerSkills::DribbleTo(const Point& pos)
{
	SKILLS_DBGN("%i: ->dribbling to position", world->state->GetSimTime());
	SKILLS_DBGN("%i: ->pos (%3.2f, %3.2f)", world->state->GetSimTime(), pos.GetX(), pos.GetY());

	Vector path = pos - world->ball->pos;
	float vzdial = path.GetLength();

	if(vzdial > DRIBBLE_STOP_DIST)
	{
		Point ciel = world->ball->pos + path * (DRIBBLE_STOP_DIST / vzdial);
		PassBallTo(ciel, 0);
	}
	else if(vzdial > BALL_POS_TOL)
	{
		PassBallTo(pos, 0);
	}
}

// 2. sposob driblovania
// POZN.: Dalsia divne naprogramovana funkcia, ktorej by som sa mal vyhnut. 
void PlayerSkills::DribbleTo2(const Point& pos, float angle)
{
	SKILLS_DBGN("%i: ->dribbling (2) to position", world->state->GetSimTime());
	SKILLS_DBGN("%i: ->pos (%3.2f, %3.2f) angle %3.2f", world->state->GetSimTime(), pos.GetX(), pos.GetY(), angle);
 
	if(angle > 90)
	{
		angle = 90;
	}
	else if(angle < -90)
	{
		angle = -90;
	}

	Point my_nextpos = world->me->pos + world->me->vel;
	Point ball_nextpos = world->ball->pos + world->ball->vel;

	float uhol = CalcDir(pos - my_nextpos, world->me->GetAbsDir()).GetDeg();

	// DASHPOWER MANAZMENT:
	// --------------------
	float dashpower;
		
	// Bezi sa co narychlejsie, ak driblujeme hlboko v obrane supera alebo nasej obrane
	if(IsPointInRectArea(world->ball->pos.GetX(), world->ball->pos.GetY(), 32.0, -34.0, 52.5, 34.0) ||
		IsPointInRectArea(world->ball->pos.GetX(), world->ball->pos.GetY(), -52.5, -34.0, -25.0, 34.0))
	{
		
		dashpower = RecommendDashToMaxSpeed();
	}
	else if(ePlayerList->NoEnemiesAround(world->me->pos))
	{
		dashpower = RecommendDashToMaxSpeed();
	}	
	else
		dashpower = RecommDashPower(true);

	// --------------------

	const float acc = min(world->params->server->player_accel_max, dashpower * world->me->type->dash_power_rate * world->me->GetEffort());

	const float karea = world->params->server->ball_size + world->me->type->player_size + world->me->type->kickable_margin;
	const float controll_dist = world->params->server->ball_size + world->me->type->player_size + 0.45f * world->me->type->kickable_margin;

	// Musim zmenit smer
	if(fabsf(uhol) > 10) 
	{
		if((ball_nextpos - my_nextpos).GetLength() < karea) // mozem sa otocit, lopta daleko neujde
		{
			float moment = CompMoment(uhol);
			m_pCommandGen->Turn(moment);
			TurnNeckTo(ball_nextpos,my_nextpos,CompTurn(moment));
		}

		// Musim loptu spracovat, aby neutiekla
		else 
		{
			Point my_targetpos = world->me->pos + world->me->vel * (1 + world->me->type->player_decay);
			float dir = (world->me->GetAbsDir() + DegAngle(uhol + angle)).GetRad();
			Point ciel = my_targetpos + RectVector(controll_dist * cosf(dir), controll_dist * sinf(dir));
			ball_nextpos = world->ball->pos + (ciel - world->ball->pos) / (1 + world->params->server->ball_decay);
			NearKick(ball_nextpos);
			TurnNeckTo(ball_nextpos,my_nextpos,0);
		}
	}
	// Idem s loptou rovno
	else 
	{
		// Ak sa nemozno pohybovat
		if(!DribbleRun(angle))
		{
			float dir = (world->me->GetAbsDir() + DegAngle(angle)).GetRad();
			uhol = world->me->GetAbsDir().GetRad();
			Point my_targetpos = world->me->pos + world->me->vel * (1 + world->me->type->player_decay) + RectVector(acc * cosf(uhol), acc * sinf(uhol));
			Point ciel = my_targetpos + RectVector(controll_dist * cosf(dir), controll_dist * sinf(dir));
			ball_nextpos = world->ball->pos + (ciel - world->ball->pos) / (1 + world->params->server->ball_decay);
			NearKick(ball_nextpos);
			TurnNeckTo(ball_nextpos,my_nextpos,0);
		}
	}
}


// KOPACIE METODY
// --------------

void PlayerSkills::ClearBall()
{
	DumbKick(Point(52.5, 0));
}

// Vykopnutie lopty celou silou na ciel

void PlayerSkills::DumbKick(const Point& target)
{
	SKILLS_DBGN("%i: ->dumb kicking to position", world->state->GetSimTime());
	SKILLS_DBGN("%i: ->pos (%3.2f, %3.2f)", world->state->GetSimTime(), target.GetX(), target.GetY());

	static float ekp, dir_diff, dist_ball;
	ekp = world->params->server->maxpower * world->params->server->kick_power_rate;
	dir_diff = ((world->ball->pos - world->me->pos) - world->me->GetAbsDir()).GetAngle().GetAbsDeg();
	dist_ball = (world->ball->pos - world->me->pos).GetLength() - world->params->server->ball_size - world->params->server->player_size;
	ekp *= 1.f - 0.25f * dir_diff / 180.f - 0.25f * dist_ball / world->me->type->kickable_margin;

	Point T1;
	if(GetLineCircleIntersect(world->ball->pos, target, world->ball->pos+world->ball->vel, ekp, T1) == 0)
	{
		m_pCommandGen->Kick(world->params->server->maxpower, CalcDir(target - world->ball->pos, world->me->GetAbsDir()).GetDeg());
		TurnTo(target);
	}
	else
	{
		TXTN("%i: ->dumb kicking from (%.2f, %.2f) to position (%.2f, %.2f)", world->state->GetSimTime(), world->ball->pos.GetX(), world->ball->pos.GetY(), target.GetX(), target.GetY());
		TXTN("  (%.2f, %.2f) => newpos (%.2f, %.2f)", world->ball->vel.GetX(), world->ball->vel.GetY(), 
			T1.GetX(), T1.GetY());
		m_pCommandGen->Kick(world->params->server->maxpower, CalcDir((T1 - (world->ball->pos+world->ball->vel)), world->me->GetAbsDir()).GetDeg());
		TurnTo(T1);
	}
}

// Vykopnutie na branku 2 (presnejsie do stredu branky)
void PlayerSkills::KickToGoal2()
{
	SKILLS_DBGN("%i: ->kicking to goal (2)", world->state->GetSimTime());

	Point target(52.5f, 0);
	PassBallTo(target, world->params->server->ball_speed_max);
}

// Vykopnutie na branku 3 (lepsie)
// POZN.: Ekvivalentnu funkciu budem potrebovat pre ohodnocovaciu mapu.
void PlayerSkills::KickToGoal3()
{
	TXTN("%i: ->kicking to goal (3)", world->state->GetSimTime());

#define MAX_OBRANCOV 32
#define SEARCH_STEP 0.2f
	// Relative angles to all defenders
	Angle obrancovia[MAX_OBRANCOV];
	int obrancov = 0;
	const float halfgoal = world->params->server->goal_width / 2;
	const float sirka = halfgoal * 0.9f;

	// Default place for kicking
	float posX = 52.5f;
	float posY = 0;

	float pomY;
	int i;
	Point enemy_pos;

	// Find all defenders in my way
	for(i = 0; i < world->enemies.GetSize(); i++)
	{
		enemy_pos = world->enemies[i].pos;
		if((enemy_pos - world->me->pos).GetLength() > 3 && enemy_pos.GetX() > world->me->pos.GetX() &&
			((enemy_pos.GetY() <= world->me->pos.GetY() && enemy_pos.GetY() >= -halfgoal) ||
			(enemy_pos.GetY() >= world->me->pos.GetY() && enemy_pos.GetY() <= halfgoal)
			))
		{
			assert(obrancov < MAX_OBRANCOV);
			DBG("%d [%.2f,%2f] ", world->enemies[i].GetUniform(), enemy_pos.GetX(), enemy_pos.GetY());
			obrancovia[obrancov] = (enemy_pos - world->ball->pos).GetAngle();
			obrancov++;
		}
	}
	DBGN(" are defenders (%d) [%.2f,%.2f]", obrancov, world->me->pos.GetX(), world->me->pos.GetY());

	// Ak sa nasli nejaki obrancovia, tak pohladaj optimalne miesto pre kopnutie
	if(obrancov > 0)
	{
		float max_vaha = -1;

		for(pomY = -sirka; pomY <= sirka; pomY += SEARCH_STEP)
		{
			// Velke cislo
			float vaha = 360;
			Angle moj_uhol = (Point(posX,pomY) - world->ball->pos).GetAngle();

			for(i = 0; i < obrancov; i++)
			{
				const float uhol = (moj_uhol - obrancovia[i]).GetAbsDeg();
				if(uhol < vaha) vaha = uhol;
			}

			if(vaha > max_vaha)
			{
				max_vaha = vaha;
				posY = pomY;
			}
		}
	}

//	PassBallTo(Point(posX, posY), world->params->server->ball_speed_max);
	// ak je brankar mimo, nestrielat do stredu brany, ale do stvrtiny
	if(posY == 0)
	{
//		TXTN("hack");
		posY = halfgoal / 2.f;
		if(world->me->pos.GetY() < 0) posY = -posY;
	}

	if(posY > 0)
		posY -= 0.15f;
	else if(posY < 0)
		posY += 0.15f;

	DumbKick(Point(posX, posY));
#undef SEARCH_STEP
#undef MAX_OBRANCOV
}

// Mam kopnut do lopty mimo modu PlayOn?
bool PlayerSkills::ShouldKickIn()
{
	// Pretazena metoda v PlayerTactics!
	PlayMode pm = world->state->GetPlayMode();
#ifdef GOALIE
	if(pm == PM_GoalKick_Our)
#else // GOALIE
	if(pm == PM_FreeKick_Our || pm == PM_IndFreeKick_Our || pm == PM_KickOff_Our || pm == PM_KickIn_Our ||
		pm == PM_CornerKick_Our || pm == PM_PenaltyKick_Our)
#endif // GOALIE
		return true;
	return false;
}

// Je mozne kopnut do lopty? (Malo by sa nad tym vobec uvazovat z hladiska sucasneho herneho modu?)
bool PlayerSkills::SomeoneShouldKickIn()
{
	// Je to jeden z nasich mod, ked zahravame?
	PlayMode pm = world->state->GetPlayMode();
	
	if(pm == PM_GoalKick_Our || pm == PM_FreeKick_Our || pm == PM_IndFreeKick_Our || pm == PM_KickOff_Our ||
		pm == PM_KickIn_Our || pm == PM_CornerKick_Our || pm == PM_PenaltyKick_Our)
		return true;
	else
		return false;
}


// KRKOVE METODY
// --------------	

// Zarovnaj krk s telom (relativny uhol medzi krkom a telom je 0 stupnov)
void PlayerSkills::AlignNeck()
{
	if(world->me->GetNeckDir().GetDeg() != 0)
	{
		m_pCommandGen->TurnNeck((int) (-world->me->GetNeckDir().GetDeg()));
		m_iTimeOfLTurnNeck = world->state->GetSimTime();
	}
}

void PlayerSkills::NeckSynchronizationWithVI(int nMode)
{
	switch(nMode)
	{
		case ALIG_NECK:
		{
			AlignNeck();

			world->me->bSendTurnNeckTo = false;
			bAvoidNextCorrection = true;
			nChangeView = 0;
			return;
		}

		case NAR_VIEW:
		{
			if(world->me->bNeckSynchronization == true)
			{
				nChangeView = 2;
				bAvoidNextCorrection = true;
			}
			else
				m_pCommandGen->ChangeView(NARROW_VIEW);

			return;	
		}

		// Prisiel prikaz na vykonanie jednej z krkovych metod (na zaklade najnovsej VI)
		case NEW_VI:
		{
			// Funkcia AlignNeck ma vacsiu prioritu ako TurnNeckTo
			
			// Korekcia otocenia krku
			float fCorrection;	

			switch(world->me->GetViewWidth())
			{
				case NARROW_VIEW:
					fCorrection = 45;
					break;

				case NORMAL_VIEW:
					fCorrection = 90;
					break;

				case WIDE_VIEW:
					fCorrection = 180;
					break;
			}

			if(bAvoidNextCorrection == false)
			{
				// Korekcia - k uhlu pridava, +/-1 * (SirkaPohladu / 3)
				fCorrection = nNeckDirection * (fCorrection / 3.0);
				// Zmenime polaritu natocenia krku
				nNeckDirection  *= -1;
			}
			else
			{
				fCorrection = 0.0;
			}

			TurnNeckToAux(fCorrection);

			if(nChangeView == 1)
			{
				m_pCommandGen->ChangeView(NORMAL_VIEW);
			}

			else if(nChangeView == 2)
			{
				m_pCommandGen->ChangeView(NARROW_VIEW);
			}
			
			world->me->bSendTurnNeckTo = false;
			bAvoidNextCorrection = false;
			nChangeView = 0;

			return;
		}

		case NOR_VIEW:
		{
			if(world->me->bNeckSynchronization == true)
			{
				nChangeView = 1;
				bAvoidNextCorrection = true;
			}
			else
				m_pCommandGen->ChangeView(NORMAL_VIEW);

			return;
		}
	
		default:
			break;
	}

}

// Tato funkcia si zaznamena udaje pre otocenie
void PlayerSkills::TurnNeckTo(const Point& future_pos, const Point& future_mypos, float nextturn)
{
	TurnNeckTo_futurePos = future_pos;
	TurnNeckTo_futureMyPos = future_mypos;
	TurnNeckTo_nextTurn = nextturn;

	// Ak je synchronizacia krku s VI vypnuta
	if(world->me->bNeckSynchronization == false)
	{
		bAvoidNextCorrection = true;
		NeckSynchronizationWithVI(NEW_VI);
	}
	
}


// Ak treba otocit krkom tak sa zavola tato funkcia
void PlayerSkills::TurnNeckToAux(float fCorrection)	
{
		// Kam treba krk natocit
		int rel = f2int(CalcDir(TurnNeckTo_futurePos - TurnNeckTo_futureMyPos, world->me->GetAbsDir() + DegAngle(TurnNeckTo_nextTurn)).GetDeg() + fCorrection);
		if(rel < world->params->server->minneckang)
		{
			rel = f2int(world->params->server->minneckang);
		}
		else if(rel > world->params->server->maxneckang)
		{
			rel = int(world->params->server->maxneckang);
		}

		int turn = f2int(rel - world->me->GetNeckDir().GetDeg());
				
		if(turn > world->params->server->maxneckmoment)
		{
			turn = int(world->params->server->maxneckmoment);
		}
		else if(turn < world->params->server->minneckmoment)
		{
			turn = f2int(world->params->server->minneckmoment);
		}

		m_pCommandGen->TurnNeck(turn);
		m_iTimeOfLTurnNeck = world->state->GetSimTime();
}


// NAHRAVACIE METODY
// -----------------

// Zadana rychlost sa prekalkuluje tak, aby bola podla vzorca brana do uvahy aj vzidalenost a natocenie hraca,
	// ked prihrava
	float PlayerSkills::CorrectPassSpeed(float fRequirSpeed)
	{	
		static float fEp, fDirDiff, fDistBball;
		fEp = world->params->server->maxpower * world->params->server->kick_power_rate;
		fDirDiff = ((world->ball->pos - world->me->pos) - world->me->GetAbsDir()).GetAngle().GetAbsDeg();
		fDistBball = (world->ball->pos - world->me->pos).GetLength() - world->params->server->ball_size - world->params->server->player_size;
		fEp *= (1.f - (0.25f * fDirDiff / 180.f) - (0.25f * fDistBball / world->me->type->kickable_margin));

		return (fRequirSpeed / fEp) * fRequirSpeed;
	}

// Nahra loptu na urcite miesto
// POZN.: Nie som si isty 100% funkcnostou tejto metody. Radsej si ju odsimulujem a naprogramujem sam. 
void PlayerSkills::PassBallTo(const Point& pos, float spd /* = 0.7f */)
{
	float tmpSpeed = spd;
	if(IsPointInRectArea(pos.GetX(), pos.GetY(), -32, -25, 52, 25))
		tmpSpeed *= 1.25f;

	SKILLS_DBGN("%i: ->passing to position", world->state->GetSimTime());
	SKILLS_DBGN("%i: ->pos (%3.2f, %3.2f)", world->state->GetSimTime(), pos.GetX(), pos.GetY());

	Vector path = pos - world->ball->pos;

	float distance = path.GetLength();
	float realSpd = distance * (1 - world->params->server->ball_decay) + tmpSpeed;
	
	if(realSpd > world->params->server->ball_speed_max)
	{
		realSpd = world->params->server->ball_speed_max;
	}

	Point target(world->ball->pos);
	if(distance > 0)
		target = world->ball->pos + path * (realSpd / distance);
/*
	ofstream out("Normalne prihavanie.txt", ios::app); // Zapis do suboru
	double lfDistance = sqrt(pow((pos.GetX() - target.GetX()), 2) + pow((pos.GetY() - target.GetY()), 2));
	out << "(" << pos.GetX() << ", " << pos.GetY() << ") => (" << target.GetX() << ", " << target.GetY() << ")" << endl;
	out << "Rozdiel: " << lfDistance << endl;
	out.close();
*/
	NearKick(target);
	TurnNeckTo(target, world->me->pos + world->me->vel, 0);
}

// Nahra loptu hracovi
// POZN.: Zase divna konstanta (.2f) vo funkcii.
void PlayerSkills::PassBallToPlayer(const Player& player, float spd /* = 0.7f */)
{
	SKILLS_DBGN("%i: ->passing to %s player %i", world->state->GetSimTime(), player.GetSide() == OUR ? "our" : (player.GetSide() == THEIR ? ">>their<<" : ">>unknown<<"), player.GetUniform());
	SKILLS_DBGN("%i: ->pos (%3.2f, %3.2f)", world->state->GetSimTime(), player.pos.GetX(), player.pos.GetY());
	
	// Ta X-ova pozicia sa mi nezda. Ak simuluje urcitu chybu, tak preco nesimuluje aj Y-ovej suradnici?
	PassBallTo(Point(player.pos.GetX() + (world->me->pos - player.pos).GetLength() * .2f, player.pos.GetY()), spd);
}

// SMYKACKOVE METODY
// -----------------
// Dokazeme smykackou vypichnut loptu?
bool PlayerSkills::CanTackleBall() const
{
	// V tomto mode sa nemoze smykat
	if(world->state->GetPlayMode() != PM_PlayOn) return false;

	// Lopta musi byt v obdlzniku sirky (2*tackle_width) a vysky (tackle_dist + tackle_back_dist)
	// obdlznik je natoceny v smere natocenia hraca
	Vector player2ball = (world->ball->pos - world->me->pos);
	player2ball.SetAngle(player2ball.GetAngle() - world->me->GetAbsDir());

	// Hrac teda moze urobit smykacku ak:
	// jr lopta vedla neho tackle_width a (za nim tackle_back_dist alebo pred nim tackle_dist)
	if(fabsf(player2ball.GetY()) >= world->params->server->tackle_width ||
	   player2ball.GetX() >= world->params->server->tackle_dist ||
	   player2ball.GetX() <= world->params->server->tackle_back_dist)
	return false;

	/*
	DBGN("Tackle can be taken v(%.2f,%.2f) p(%.2f,%.2f;%.2f) b(%.2f,%.2f)",
		player2ball.GetX(), player2ball.GetY(),
		world->me->pos.GetX(), world->me->pos.GetY(), world->me->GetAbsDir().GetAbsDeg(),
		world->ball->pos.GetX(), world->ball->pos.GetY());
	*/
	return true;
}
// Urob smykacku
void PlayerSkills::Tackle(const float power)
{
	m_pCommandGen->Tackle(power);
}


// ZACHYTAVACIE METODY
// -------------------

// Zisti, ci lopta je v danom cykle na dosah hraca
// Validne hodnoty pre cas su: 0,1,2,3,...
// pre lfBallDecayInTime = pow(ball_decay, time (cas 0 je specialny pripad)
bool PlayerSkills::IsBallInPlayerKickableAreaDuringInteception(double lfBallDecayInTime, int time)
{
	// Toto je hracov dosah, kedy este moze kopnut do lopty	
	const double lfPlayerRadius = world->params->server->ball_size + world->me->type->player_size + world->me->type->kickable_margin * 0.51;
	
	// Specialny pripad, neobsahuje geometrice rady
	if (time == 0 && lfPlayerRadius >= 0)
		return (world->ball->pos - world->me->pos).GetLength() <= (lfPlayerRadius * lfPlayerRadius);
	
	// Tento vektor je vektor od hraca (v case 0) k lopte v case "time"
	Vector vectFromPlayerToBall = world->ball->pos + (world->ball->vel * (1.0 - lfBallDecayInTime)/(1.0 - world->params->server->ball_decay)) - world->me->pos;

	// Optimisticky odhad
//	double lfPlayerCircleRadius = time * world->me->type->player_speed_max + lfPlayerRadius;
	
	// Realnejsi odhad
	double lfPlayerCircleRadius = time * (world->me->vel).GetLength() + lfPlayerRadius;


/* ----------------------- LOG ----------------------- //
	char buffer[256];
	sprintf(buffer, "!TXT Interception\\%d", world->me->GetUniform());
	ofstream log1(buffer, ios::app);
	
	log1 << "\n  Cas = " << time 
		<< "\n  Vzdialenost = " << vectFromPlayerToBall.GetLength()
		<< "\n  Okruh hraca = " << time * world->me->type->player_speed_max
		<< "\n  Pozicia lopty[" << time <<"] = " << (vectFromPlayerToBall + world->me->pos).GetX() << ", " << (vectFromPlayerToBall + world->me->pos).GetY() << endl;

	log1.close();
// ----------------------- // LOG ----------------------- */

  return (vectFromPlayerToBall.GetLength() <= lfPlayerCircleRadius);
}

// Vypocita optimalny bod pre zachytenie lopty
Point PlayerSkills::GetBallInterceptionPoint()
{
	const int MAX_STEPS = 14; // 2^(14-2) = 4096 > 3000	
	  
	double gamma[MAX_STEPS];
	int time[MAX_STEPS];
	
	int nResTime;
	Point pointResPosition;

	time[0]		= 0;
	gamma[0]	= 1.0;
	
	time[1] = 1;
	gamma[1] = world->params->server->ball_decay;
	
	if (IsBallInPlayerKickableAreaDuringInteception(gamma[0], time[0]))
	{
		nResTime = time[0];
		pointResPosition = world->ball->pos;
		return pointResPosition;
	} 

	if (IsBallInPlayerKickableAreaDuringInteception(gamma[1], time[1])) 
	{
		nResTime = time[1];
		pointResPosition = world->ball->pos + world->ball->vel;
		return pointResPosition;
	}

	int i = 2;
  
	while(1)
	{
		gamma[i] = gamma[i-1] * gamma[i-1];
		time[i] = time[i-1] + time[i-1];
	  
		if (IsBallInPlayerKickableAreaDuringInteception(gamma[i], time[i])) 
			break;
		i++;

		if ( i >= MAX_STEPS )
		{
			// Potrebujem viac ako time[i-1] krokov 
			nResTime= time[i-1] ;
			pointResPosition =  world->ball->pos +  world->ball->vel * (1.0 - gamma[i-1]) / ( 1.0 - world->params->server->ball_decay);
			return pointResPosition;
		}
	}
	
	i = i - 1;
  
	int nMinTime = time[i];
	int nMidTime = time[i] + time[i-1];
	int nMaxTime = time[i+1];
	double lfMinGamma = gamma[i];
	double lfMidGamma = gamma[i] * gamma[i-1];
	double lfMaxGamma = gamma[i+1];

	while (i > 1)
	{
		if (IsBallInPlayerKickableAreaDuringInteception(lfMidGamma, nMidTime))
		{
		  lfMaxGamma = lfMidGamma;
		  lfMidGamma = lfMinGamma * gamma[i-2];
		  nMaxTime = nMidTime;
		  nMidTime = nMinTime + time[i-2];
		}
		else
		{
		  lfMinGamma = lfMidGamma;
		  lfMidGamma = lfMinGamma * gamma[i-2];
		  nMinTime = nMidTime;
		  nMidTime = nMinTime + time[i-2];
		}
 
		i= i-1;
	}

	nResTime = nMaxTime;  
	pointResPosition = world->ball->pos + world->ball->vel * (1.0 - lfMaxGamma)/(1.0 - world->params->server->ball_decay);

/* ----------------------- LOG ----------------------- //
	char buffer[256];
	sprintf(buffer, "!TXT Interception\\%d", world->me->GetUniform());
	ofstream log1(buffer, ios::app);
	int nSimTime = world->state->GetSimTime();


	log1 << "Vysledny cas: " << nResTime << endl
	<< "Prerusovaci bod: (" << pointResPosition.GetX() << ", " << pointResPosition.GetY() << ")" << endl;
	log1.close();
// ----------------------- // LOG ----------------------- */
	
	return pointResPosition;	
}

// Zachyti loptu ak je to mozne. Vracia false ak to nie je mozne.
bool PlayerSkills::InterceptBall()
{
	if(!DependentForBall())
		return false;

	Point pointInterception = GetBallInterceptionPoint();
	Vector vectPlayerFirstPossibleInterceptionPoint = pointInterception - world->me->pos;

	Vector vectFromPlayerToBall;
	double lfMinDistance = vectPlayerFirstPossibleInterceptionPoint.GetLength();
	double lfActualDistance = lfMinDistance;

	bool bTmp = false;

	// Tento vektor je vektor od hraca (v case i) k lopte v case "time"
	for(int i = 1; i <= 3000; i++)
	{
		vectFromPlayerToBall = world->ball->pos + (world->ball->vel * (1.0 - i)/(1.0 - world->params->server->ball_decay)) - world->me->pos;
		lfActualDistance = vectFromPlayerToBall.GetLength();
		if(lfActualDistance < lfMinDistance)
		{
			lfMinDistance = lfActualDistance;
			bTmp = true;
		}
		else if(bTmp == true)
			break;
	}

	if(bTmp == true)
	{
		pointInterception = world->me->pos + vectFromPlayerToBall;
	}
	
	// Optimalne miesto, kde zachytime loptu
	RunToBallInteceptPosition(pointInterception);
		return true;
}


// Mal by hrac vobec uvazovat nad zachytenim lopty?
bool PlayerSkills::ShouldInterceptBall()
{
	if(world->state->GetPlayMode() != PM_PlayOn)
	{
		if(ShouldKickIn())
			return true;
		else
			return false;
	}

	return true;
}


// Mal by som bezat za loptou?
bool PlayerSkills::DependentForBall() const
{
	int ourc;
	int theirc;
	int closestFriend = NearFriend(&ourc);
	int enemy = NearEnemy(&theirc);

	if(closestFriend == -1)
		return true;
	else
	{
		// Nas hrac tam bude skor
		if(enemy == -1 || theirc > ourc)
			return false;
		else
		{
			// Medzi nasim hracom a loptou je super
			if(InTheMiddle(world->enemies[enemy].pos, world->friends[closestFriend].pos, world->ball->pos))
			{
				// Najblizsi super nie je medzi mnou a loptou
				bool dependent = !InTheMiddle(world->enemies[enemy].pos, world->me->pos, world->ball->pos);
				
				// Nikto od nas nie je medzi mnou a loptou
				for (int i = 0; i < world->friends.GetSize() && dependent; i++)
				{
					if(InTheMiddle(world->friends[i].pos, world->me->pos, world->ball->pos))
						dependent = false;
				}

				return dependent;
			}
			else
				return false;
		}
	}
}

// Vrati supera najblizsieho k lopte
int PlayerSkills::NearEnemy(int* pCycles) const
{
	if(world->enemies.GetSize() == 0)
		return -1;

	float psteps;
	Ball tmpBall(*world->ball);
	Vector diff;
	const float karea = world->params->server->player_size + world->params->server->ball_size + world->params->server->kickable_margin;

	for(int ncycles = 1; ; ncycles++)
	{
		tmpBall.Step();

		for(int i = 0; i < world->enemies.GetSize(); i++)
		{
			diff = tmpBall.pos - world->enemies[i].pos;

			psteps = (diff.GetLength() - karea) / world->params->server->player_speed_max;
			if(psteps < ncycles)
			{
				*pCycles = ncycles;
				return i;
			}
		}
	}
}

// Vrati naseho najblizsieho hraca k lopte. Vrati -1, ak som to ja.
int PlayerSkills::NearFriend(int* pCycles) const
{
	float psteps;
	Ball tmpBall(*world->ball);
	Vector diff;

	const float mykarea = world->me->type->player_size + world->params->server->ball_size + world->me->type->kickable_margin;
	
	float kareas[12];
	float speeds[12];
	for(unsigned k = 0; k < 12; k++)
	{
		kareas[k] = world->params->server->ball_size + world->friendTypes[k]->player_size + world->friendTypes[k]->kickable_margin;
		speeds[k] = world->friendTypes[k]->player_speed_max;
	}

	for(int ncycles = 1; ; ncycles++)
	{
		tmpBall.Step();
		diff = tmpBall.pos - world->me->pos;

		psteps = (diff.GetLength() - mykarea) / world->me->type->player_speed_max;
		if(psteps < ncycles)
		{
			*pCycles = ncycles;
			return -1;
		}

		for(int i=0;i < world->friends.GetSize();i++)
		{
			diff = tmpBall.pos - world->friends[i].pos;

			psteps = (diff.GetLength() - kareas[world->friends[i].GetUniform()])/ speeds[world->friends[i].GetUniform()];
			if(psteps < ncycles)
			{
				*pCycles = ncycles;
				return i;
			}
		}
	}
}

// ZVYSNE METODY
// -------------
	
// Rozhodovanie sa
void PlayerSkills::Act()
{
#ifdef GOALIE
	world->me->bNeckSynchronization = false;
#endif

	m_BallDiff = world->ball->pos - world->me->pos;
	m_fBallDistance = m_BallDiff.GetLength();
	
	if(m_fBallDistance < (world->me->type->player_size + world->params->server->ball_size) )
	{
		m_fBallDistance = world->me->type->player_size + world->params->server->ball_size;
	}

	if(world->state->IsTimeStoped()) 
	{
		m_bKickable = false;
		RunPlayOff();
	}
	else
	{
//		TXTN("%d: Acting", world->state->GetSimTime());
		m_bKickable = m_fBallDistance < world->params->server->ball_size + world->me->type->player_size + world->me->type->kickable_margin;
		RunPlayOn();
	}
 	if(m_iGoalieCatchBan > 0)
		m_iGoalieCatchBan--;

	if((world->me->bNeckSynchronization) && (world->me->bSendTurnNeckTo))
		NeckSynchronizationWithVI(NEW_VI);
}

#ifndef GOALIE 
// Vrati poziciu ofsajdovej ciary
// POZN.: Na co akoze ma sluzit premenna fast?
float PlayerSkills::GetOffsideLine() const
{
	float offside_line = 0;
	float fast = 0;
	float second = 0;
	float tmp;
	int i;

	for (i = 0; i < world->enemies.GetSize(); i++)
	{	// Druhy obranca
		if (world->enemies[i].pos.GetX() > second)
		{
			second = world->enemies[i].pos.GetX();
			
			if (second > fast)
			{
				tmp = fast;
				fast = second;
				second = tmp;
			}
		}
	}

	if (second > world->ball->pos.GetX())
	{
		offside_line = second;
	}
	else
	{
		offside_line = world->ball->pos.GetX();
	}

	return offside_line;
}

#endif

// Je dany bod medzi bodmi p1 a p2?
bool PlayerSkills::InTheMiddle(const Point& tested, const Point& p1, const Point& p2) const
{
	const float distP1P2 = (p1 - p2).GetLength();

	if ((tested - p1).GetLength() > distP1P2)
		return false;
	if ((tested - p2).GetLength() > distP1P2)
		return false;

	return true;
}

// Predikuje rychlost v dalsom kroku
Vector PlayerSkills::PredictNextVel(float DashPower) const
{
	if(DashPower > world->params->server->maxpower)
	{
		DashPower = world->params->server->maxpower;
	}
	else if(DashPower < world->params->server->minpower)
	{
		DashPower = world->params->server->minpower;
	}

	float Ep = DashPower * world->me->type->dash_power_rate * world->me->GetEffort();
	
	if(Ep > world->params->server->player_accel_max)
	{
		Ep = world->params->server->player_accel_max;
	}
	else if(Ep < -world->params->server->player_accel_max)
	{
		Ep = -world->params->server->player_accel_max;
	}

	float uhol = world->me->GetAbsDir().GetRad();
	
	RectVector nextVel(world->me->vel.GetX() + Ep * cosf(uhol),world->me->vel.GetY() + Ep * sinf(uhol));

	if(nextVel.GetLength() > world->me->type->player_speed_max)
	{
		nextVel *= (world->me->type->player_speed_max / nextVel.GetLength());
	}

	return nextVel;
}

// Otacaj sa zatial co hladas loptu, alebo nieco ine.
float PlayerSkills::Scan()
{
	float uhol = 0;
	const float TURNFACTOR = 0.875f;
	
	NeckSynchronizationWithVI(ALIG_NECK);

	switch(world->me->GetViewWidth())
	{
	case NARROW_VIEW:
		
	NeckSynchronizationWithVI(NOR_VIEW);
	// tu nie je break naschval

	case NORMAL_VIEW:
		uhol = world->params->server->visible_angle * TURNFACTOR;
		break;

	case WIDE_VIEW:
		uhol = world->params->server->visible_angle * TURNFACTOR * 2.0f;
		break;
	}

	m_pCommandGen->Turn(CompMoment(uhol));
	SetNextSynchMode(Synch_WaitForVI);

	return uhol;
}

// Zmeni uhol pohladu
void PlayerSkills::SetView(const ViewWidth vw)
{
	m_pCommandGen->ChangeView(vw);
}

// Kopnutim sa natoci lopta
void PlayerSkills::TurnBall()
{
	SKILLS_DBGN("%i: ->turn ball", world->state->GetSimTime());

	// Ak je dostatocne blizko
	if(IsBallKickAble())
	{
		Angle target_ang = m_BallDiff.GetAngle() + DegAngle(0.4f * 180.f);
		const float controll_dist = world->me->type->player_size + world->params->server->ball_size + 0.45f * world->me->type->kickable_margin;
		Vector rel_target = PolarVector(controll_dist, target_ang);

		// Miesto pozadovaneho kopnutia
		Point ciel = (world->me->pos + world->me->vel + rel_target);
		NearKick(ciel);
	}
	else // Nemozeme kopnut
	{
		return;
	}
}

// Natoci telo na urcity bod
void PlayerSkills::TurnTo(const Point& pos)
{
	PlayMode pm = world->state->GetPlayMode();

	if(!world->ball->IsValid() || (pm == PM_GoalKick_Our || pm == PM_FreeKick_Our || pm == PM_IndFreeKick_Our ||
		pm == PM_KickOff_Our || pm == PM_KickIn_Our || pm == PM_CornerKick_Our || pm == PM_PenaltyKick_Our))
	{			
		TurnNeckTo_futurePos = pos;
		TurnNeckTo_futureMyPos = world->me->pos;
		TurnNeckTo_nextTurn = 0;
	}

	Angle dir = CalcDir(pos - world->me->pos, world->me->GetAbsDir());
	if(dir.GetAbsDeg() >= 1)
		m_pCommandGen->Turn(dir.GetDeg());
}


// POSIELANIE SPRAV
// ----------------
// Posiela zvukovu informaciu
// POZN.: Neoptimalne vyuzita dlzka spravy. Okrem toho najdolezitejsia je pozicia hraca, ostatne parametre hraca
// nie je podla mna potrebne prenasat.
void PlayerSkills::SendInfo()
{
	// Mal by som nieco povedat?
	if(!ShouldSendInfo())
		return;
	
	// Vytvori spravu. Uklada udaje do 3-stavovych a 2-stavovych premennych. Kombinuje premenne (2x 3-stavovu a
	// 3x dvojstavovu) do cisla z intervalu <0, 71>. Prekonvertuje cisla do validnych znakov (okrem '_' a ' ')
	// Pozri aj PlayerSkills::OnHear()

	unsigned char parts[10][6];
	int i;
	for(i = 0; i < 10; i++)
	{
		parts[i][0] = 0xff;
		parts[i][1] = parts[i][2] = parts[i][3] = parts[i][4] = parts[i][5] = 0;
	}

	// Alokacia casti:
	// - = volne, X = hracova x-ova suradnica, Y = hracova y-ova suradnica
	// D = hracov smer, S = hracova rychlost
	// 1(3) XYDS----YY
	// 2(3) XYDS----DX
	// 3(2) XYDS----DX
	// 4(2) XYDS----SX
	// 5(2) XYDS----YX
	// Toto musi byt to iste v OnHear().

//	DBGN("---");
//	DBGN("SEND INFO: %0.3f, %0.3f, %0.3f, %0.3f", world->me->pos.GetX(), world->me->pos.GetY(), world->me->vel.GetX(), world->me->vel.GetY());

	// X suradnica hraca (vyuziva 72,3,2,2,2 pozri hore)
	int tmp = int((world->me->pos.GetX() - world->params->flags[L0].GetX()) * float(72 * 3 * 2 * 2 * 2) / 
		(world->params->flags[R0].GetX() - world->params->flags[L0].GetX()) + 0.5);
//	DBGN("tmp_x %i", tmp);
	parts[9][5] = tmp % 2; tmp /= 2;
	parts[9][4] = tmp % 2; tmp /= 2;
	parts[9][3] = tmp % 2; tmp /= 2;
	parts[9][2] = tmp % 3; tmp /= 3;
	parts[0][0] = tmp % 72; tmp /= 72;

	// Y suradnica hraca (vyuziva 72,3,3,2 pozri hore)
	tmp = int((world->me->pos.GetY() - world->params->flags[T0].GetY()) * float(72 * 3 * 3 * 2) / 
		(world->params->flags[B0].GetY() - world->params->flags[T0].GetY()) + 0.5);
//	DBGN("tmp_y %i", tmp);
	parts[8][5] = tmp % 2; tmp /= 2;
	parts[8][1] = tmp % 3; tmp /= 3;
	parts[9][1] = tmp % 3; tmp /= 3;
	parts[1][0] = tmp % 72; tmp /= 72;

	// Smer hraca (vyuziva 72,3,2 pozri hore)
	float angle = world->me->vel.GetAngle().GetRad();
	while(angle < 0)
		angle += 2.0f * M_PI; 
	while(angle > 2.0f * M_PI)
		angle -= 2.0f * M_PI;
	
	tmp = int(angle * float(72 * 3 * 2) / (2.f * M_PI) + 0.5);
//	DBGN("tmp_dir %i ang %0.3f", tmp, angle);
	parts[8][3] = tmp % 2; tmp /= 2;
	parts[8][2] = tmp % 3; tmp /= 3;
	parts[2][0] = tmp % 72; tmp /= 72;

	// Rychlost hraca (vyuziva 72,2 pozri hore)
	tmp = int(world->me->vel.GetLength() * float(72 * 2) /
		(world->params->server->player_speed_max + world->params->player->player_speed_max_delta_max) + 0.5);
//	DBGN("tmp_spd %i", tmp);
	parts[8][4] = tmp % 2; tmp /= 2;
	parts[3][0] = tmp % 72; tmp /= 72;

	//world->ball->pos.GetX(), world->ball->pos.GetY(), world->ball->vel.GetX(), world->ball->vel.GetY(), ball.age);

	// Prekonvertuje casti do cisiel a zakoduje ich do znakov
	char msg[11];
	memset(msg, 0, 11);
	for(i = 0; i < 10; i++)
	{
		if(parts[i][0] == 0xff)
			parts[i][0] = (((parts[i][5] * 2 + parts[i][4]) * 2 + parts[i][3]) * 3 + parts[i][2]) * 3 + parts[i][1];

		if((parts[i][0] >= 0) && (parts[i][0] <= 25))
			msg[i] = 'a' + char(parts[i][0]);
		else if((parts[i][0] >= 26) && (parts[i][0] <= 51))
			msg[i] = 'A' + char(parts[i][0] - 26);
		else if((parts[i][0] >= 52) && (parts[i][0] <= 61))
			msg[i] = '0' + char(parts[i][0] - 52);
		else if(parts[i][0] == 62)
			msg[i] = '+'; 
		else if(parts[i][0] == 63) 
			msg[i] = '-';
		else if(parts[i][0] == 64)
			msg[i] = '*'; 
		else if(parts[i][0] == 65)
			msg[i] = '/';
		else if(parts[i][0] == 66)
			msg[i] = '('; 
		else if(parts[i][0] == 67)
			msg[i] = ')';
		else if(parts[i][0] == 68)
			msg[i] = '.'; 
		else if(parts[i][0] == 69)
			msg[i] = '?';
		else if(parts[i][0] == 70)
			msg[i] = '<'; 
		else if(parts[i][0] == 71) 
			msg[i] = '>';
		else
		{ 
			WARNN("failed encoding for say (%i->%i)", i, (int) parts[i][0]);
			return;
		}
	}

/*	for(i = 0; i < 10; i++) DBG("%i ", (int) parts[i][0]); DBGN("");
	for(i = 0; i < 10; i++) DBG("%i ", (int) parts[i][1]); DBGN("");
	for(i = 0; i < 10; i++) DBG("%i ", (int) parts[i][2]); DBGN("");
	for(i = 0; i < 10; i++) DBG("%i ", (int) parts[i][3]); DBGN("");
	for(i = 0; i < 10; i++) DBG("%i ", (int) parts[i][4]); DBGN("");
	for(i = 0; i < 10; i++) DBG("%i ", (int) parts[i][5]); DBGN("");
*/
	msg[10] = '\0';
	m_pCommandGen->Say(msg);
}

// Mal by som poslat zvukovu informaciu?
bool PlayerSkills::ShouldSendInfo()
{
	if(world->ball->pos.GetX() < 0.f)
	{
		// Lopta je na nasej polovici. Hraci 2 az 11 by mali posielat zvukove spravy kazdych 20 cyklov.
		int time = world->state->GetSimTime() % 20;
		
		if(time == (world->me->GetUniform() - 2) * 2)
			return true;
		else
			return false;
	}
	else
	{
		// Lopta je na polovici supera. Hraci na ich strane a na opacnej strane od lopty by mali kricat
		// kazde dva cykly.
		if(!world->ball->IsValid())
		{
			SHOUT_DBGN("SHT %i: invalid ball", world->state->GetSimTime());
			return false;
		}

		if(world->me->pos.GetX() < 0.f)
		{
			SHOUT_DBGN("SHT %i: in defense", world->state->GetSimTime()); return false;
		}

		if(world->me->pos.GetY() * world->ball->pos.GetY() > 0.f)
		{
			SHOUT_DBGN("SHT %i: same side", world->state->GetSimTime());
			return false;
		}

		static int lastTime = -1000;
		
		if(world->state->GetSimTime() - lastTime < 2)
		{
			SHOUT_DBGN("SHT %i: too often", world->state->GetSimTime());
			return false;
		}

		SHOUT_DBGN("SHT %i: shouting", world->state->GetSimTime());
		return true;
	}
}

// PUBLIC METODY

// Pretazena metoda z PlayerKernel
void PlayerSkills::OnInit(RealSide realSide, int uniform, PlayMode mode)
{
	PlayerKernel::OnInit(realSide, uniform, mode);
	NeckSynchronizationWithVI(NAR_VIEW);
	SetNextSynchMode(Synch_VIOnly);
}

// Pretazena metoda z PlayerKernel
// Spracuje zvukovu informaciu
void PlayerSkills::OnHear(int time, float direction, bool ourTeam, int uniform, const char* msg)
{
	// Ignoruj spravy z druheho timu
	if(!ourTeam)
		return;

	// Ignoruj tiez, prilis kratke spravy
	if(strlen(msg) != 10)
		return;

	// Zisti informacie o odosielatelovi
	Player player(world);
	player.SetUniform(uniform);

	// Parsujeme jeho spravu. Kazdy znak je skonvertovany na cislo z intervalu <0, 71> (ignorujeme '_' a ' ').
	// Cislo je rozdelene do troch casti: 2x 3-stavove a 3x 2-stavove (3*3*2*2*2=72)
	// Pouzivame tri casti a interval <0, 71> cisiel na ulozenie udajov efektivnym sposobom.
	// Pozri aj PlayerSkills:SendInfo()

	unsigned char parts[10][6];
	for(int i = 0; i < 10; i++)
	{
		// Konvertuje do <0, 71>
		unsigned char num;
		if((msg[i] >= 'a') && (msg[i] <= 'z'))
			num = ((char) msg[i]) - 'a';
		else if((msg[i] >= 'A') && (msg[i] <= 'Z'))
			num = ((char) msg[i]) - 'A' + 26;
		else if((msg[i] >= '0') && (msg[i] <= '9'))
			num = ((char) msg[i]) - '0' + 52;
		else if(msg[i] == '+')
			num = 62;
		else if(msg[i] == '-')
			num = 63;
		else if(msg[i] == '*')
			num = 64;
		else if(msg[i] == '/')
			num = 65;
		else if(msg[i] == '(')
			num = 66;
		else if(msg[i] == ')')
			num = 67;
		else if(msg[i] == '.')
			num = 68;
		else if(msg[i] == '?')
			num = 69;
		else if(msg[i] == '<')
			num = 70;
		else if(msg[i] == '>')
			num = 71;
		else
		{
			WARNN("failed decoding for hear (%i->%c)", i, msg[i]);
		return;
		}

		// Rozdelene do casti
		parts[i][0] = num;
		parts[i][1] = num % 3; num /= 3;
		parts[i][2] = num % 3; num /= 3;
		parts[i][3] = num % 2; num /= 2;
		parts[i][4] = num % 2; num /= 2;
		parts[i][5] = num % 2; num /= 2;
	}

	Point* flags = world->params->flags;

/*	DBGN("---");
	for(i = 0; i < 10; i++) DBG("%i ", (int) parts[i][0]); DBGN("");
	for(i = 0; i < 10; i++) DBG("%i ", (int) parts[i][1]); DBGN("");
	for(i = 0; i < 10; i++) DBG("%i ", (int) parts[i][2]); DBGN("");
	for(i = 0; i < 10; i++) DBG("%i ", (int) parts[i][3]); DBGN("");
	for(i = 0; i < 10; i++) DBG("%i ", (int) parts[i][4]); DBGN("");
	for(i = 0; i < 10; i++) DBG("%i ", (int) parts[i][5]); DBGN("");
*/

	// Alokacia casti:
	// - = volne, X = hracova x-ova suradnica, Y = hracova y-ova suradnica
	// D = hracova smer, S = hracova rychlost
	// 1(3) XYDS----YY
	// 2(3) XYDS----DX
	// 3(2) XYDS----DX
	// 4(2) XYDS----SX
	// 5(2) XYDS----YX
	// This has to be the same as in SendInfo().

	// X suradnica hraca (pouziva sa 72,3,2,2,2 pozri hore)
	int tmp = (((parts[0][0] * 3 + parts[9][2]) * 2 + parts[9][3]) * 2 + parts[9][4]) * 2 + parts[9][5];
//	DBGN("tmp_x %i", tmp);
	player.pos.SetX(float(tmp) * (flags[R0].GetX() - flags[L0].GetX()) / 
		float(72 * 3 * 2 * 2 * 2) + flags[L0].GetX());

	// Y suradnica hraca (pouziva sa 72,3,3,2 pozri hore)
	tmp = ((parts[1][0] * 3 + parts[9][1]) * 3 + parts[8][1]) * 2 + parts[8][5];
//	DBGN("tmp_y %i", tmp);
	player.pos.SetY(float(tmp) * (flags[B0].GetY() - flags[T0].GetY()) / 
		float(72 * 3 * 3 * 2) + flags[T0].GetY());

	// Smer hraca (pouziva sa 72,3,2 pozri hore)
	tmp = (parts[2][0] * 3 + parts[8][2]) * 2 + parts[8][3];
	float dir = float(tmp) * 2.0f * M_PI / float(72 * 3 * 2);
//	DBGN("tmp_dir %i ang %0.3f", tmp, dir);

	// Rychlost hraca (pouziva sa 72,2 pozri hore)
	tmp = parts[3][0] * 2 + parts[8][4];
//	DBGN("tmp_spd %i", tmp);
	float spd = float(tmp) * (world->params->server->player_speed_max + world->params->player->player_speed_max_delta_max) / float(72 * 2);

	player.vel.SetX(float(spd * cos(dir)));
	player.vel.SetY(float(spd * sin(dir)));
	player.SetAge(MAX_PLAYER_AGE - 10);

	// Uloz informaciu o odosielatelovi
	int index = world->FindFriendIndex(player.GetUniform());
	
	if(index == -1)
		world->friends.Add(player);
	else
		world->friends[index] = player;

	SHOUT_DBGN("SHT %i: RECV INFO: %i: %0.2f, %0.2f, %0.2f, %0.2f", world->state->GetSimTime(), player.GetUniform(), player.pos.GetX(), player.pos.GetY(), player.vel.GetX(), player.vel.GetY());
	if(index == -1) SHOUT_DBGN("SHT %i: Hmmm, new player", world->state->GetSimTime());

//	&other_ball.pos.GetX(), &other_ball.pos.GetY(), &other_ball.vel.GetX(), &other_ball.vel.GetY(), &other_ball.age,

/*	if(ball.valid) //viem o lopte, takze mozem upravit poziciu
	{
		float mydist = (world->ball->pos - world->me->pos).GetLength();
		float hisdist = (other_ball.pos - other_player.pos).GetLength();

		if(hisdist < mydist && other_ball.age < ball.age) //podmienka aktualizacie
		{
			ball = other_ball;
			ball.valid = true;
		}
	}*/
}

// Zisti, ci bod patri do obdlznikovej oblasti
bool PlayerSkills::IsPointInRectArea(double lfPointX, double lfPointY, double lfLeft, double lfTop, double lfRight, double lfBottom)
{
	if((lfLeft <= lfPointX) && (lfPointX <= lfRight) && (lfTop <= lfPointY) && (lfPointY <= lfBottom))
		return true;
	else
		return false;
}

//
float PlayerSkills::CalcRealSpeedMax()const
{
	PLAYER_TYPE* player_type = world->me->type;
	float max_accel = 100.0f * player_type->dash_power_rate * player_type->effort_max;
    float real_speed_max = max_accel / (1.0f - player_type->player_decay);
    if ( real_speed_max > player_type->player_speed_max )
    {
        real_speed_max = player_type->player_speed_max;
    }
	return real_speed_max;
}

// Kompenzovanie momentu zotrvacnosti
float PlayerSkills::CompMoment(float Moment) const
{
	float retval = Moment * (1 + world->me->vel.GetLength() * world->me->type->inertia_moment);
	if(retval > world->params->server->maxmoment)
	{
		return world->params->server->maxmoment;
	}
	else if(retval < world->params->server->minmoment)
	{
		return world->params->server->minmoment;
	}
	return retval;
}
// Kompenzovanie momentu otacania
float PlayerSkills::CompTurn(float TurnMoment) const
{
	return TurnMoment / (1 + world->me->vel.GetLength() * world->me->type->inertia_moment);
}

// Vypocita kolko krokov by trvalo prekonat danu vzdialenost so zadanym koeficientom sprintu a rychlosti
int PlayerSkills::PredictSteps(float dist, float Power, float aktSpeed) const
{
	int Steps = 0;
	const float acc = Power * world->me->GetEffort() * world->me->type->dash_power_rate;
	while(dist > 0)
	{
		aktSpeed = min(world->me->type->player_speed_max, aktSpeed + acc);
		dist -= aktSpeed;
		aktSpeed *= world->me->type->player_decay;
		Steps++;
	}

	return Steps;
}

// Vypocita doporuceny maximalny koeficient sprintu
float PlayerSkills::RecommDashPower(bool fast, bool reallyFast) const
{
	const float power_low = min((world->me->type->stamina_inc_max) * world->params->server->recover_min, world->params->server->maxpower);
	const float stamina_low = world->params->server->stamina_max * world->params->server->effort_dec_thr;
	
	if(world->me->GetEffort() < world->me->type->effort_max)
		return power_low; // je unaveny
	if(world->me->GetStamina() <= stamina_low)
		return power_low; // pod 30%
	
	const float stamina_mid = world->params->server->stamina_max * world->params->server->effort_inc_thr;
	const float slow_powermax = min(world->me->type->stamina_inc_max, world->params->server->maxpower);
	
    float result;
    
	if(fast && world->state->GetPlayMode() == PM_PlayOn) // bezime rychlo
	{
		if(world->me->GetStamina() < (stamina_mid)) // pod 60%
		{
			result = slow_powermax + (world->me->GetStamina() - stamina_low) * (world->params->server->maxpower - slow_powermax)/(stamina_mid - stamina_low);
		}
		else // nad 60%
		{
			result = world->params->server->maxpower;
		}
	}
	else // bezime pomaly
	{
		if(world->me->GetStamina() < stamina_mid) // pod 60%
		{
			result = power_low;
		}
		else // nad 60%
		{
			
			if(world->me->GetStamina() < world->params->server->stamina_max) // pod 100%
			{
				result = power_low + (world->me->GetStamina() - stamina_mid) * (slow_powermax - power_low)/(world->params->server->stamina_max - stamina_mid);
			}
			else // stamina je plna
			{
				result = slow_powermax;
			}
		}
	}

	if (reallyFast)
	{
        result *= 1.5f;
		if (result > 100.0f)
			result = 100.0f;
    }

	float MaxNeededDash = RecommendDashToMaxSpeed();
/*	DBGN("\n%d", world->state->GetSimTime());
	DBGN("\n\n\n\nRealDash %3.2f", result);
	DBGN("\n\tToMaxSpeed: %3.2f", MaxNeededDash);
*/
	if (result > MaxNeededDash)
		result = MaxNeededDash;

    return result;
}

//
float PlayerSkills::RecommendDashToMaxSpeed() const
{
	float RealSpeedMax = CalcRealSpeedMax();
	float neededAcceleration = RealSpeedMax - world->me->vel.GetLength();
//	DBGN("Needed accel: %f", neededAcceleration);
	float power = neededAcceleration / (world->me->effort * world->params->server->dash_power_rate);
	
	if (power > world->params->server->maxpower)
		power = world->params->server->maxpower;
	return power;
}

// Vrati pozorovaci uhol v zavislosti na vzdialenosti
ViewWidth PlayerSkills::WidthFromDistance(float Distance) const
{
	if(GetPlayerRole() == PR_Goalie)
	{
		if(Distance < 20)
		{
			return NARROW_VIEW;
		}
		else if(Distance > 40.0f)
		{
			return WIDE_VIEW;
		}
		else
		{
			return NORMAL_VIEW;
		}
	}
	else 
	{	
		// @JW

		PlayMode pm = world->state->GetPlayMode();

#ifndef GOALIE 
		
		if((pm == PM_FreeKick_Our || pm == PM_IndFreeKick_Our || pm == PM_KickIn_Our || pm == PM_CornerKick_Our) &&
			(Distance < 1.0f))
		{
			return NORMAL_VIEW;
		}
#else

		if((pm == PM_FreeKick_Our || pm == PM_GoalKick_Our || pm == PM_IndFreeKick_Our || pm == PM_KickIn_Our || pm == PM_CornerKick_Our) &&
			(Distance < 1.0f))
		{
			return WIDE_VIEW;
		}
#endif		
		// @JW
		else if(Distance < 15.0f) // Default 5.f)			// (0.75f * world->params->server->visible_distance)
		{
			return NARROW_VIEW;
		}
		else if(Distance > 30.0f) // Default 20.0f)
		{
			return WIDE_VIEW;
		}
		else
		{
			return NORMAL_VIEW;
		}
	}
}

/**************************************************************************************************/
/*
/*	FUNKCIE PRE NEURONOVE SIETE	(FcFarmari) - pouzivat na vlastne nebezpecie!
/*
/**************************************************************************************************/

int PlayerSkills::FCEvaluateInputs(FCNetwork *net, int N, double **Inputs)
{
	int i, Index = 0, diffTime = world->state->GetSimTime() - FCTime;
	double Res, Max = 0.0;

	for (i = 0; i < N; i++)
	{
#ifdef DRIBBLING_TESTING
		net->SetInput(Inputs[i], 0);
		Res = net->GetOutput(0);
#else
		net->SetInput(Inputs[i], (FCDiffTime));
		Res = net->GetOutput(FCDiffTime);
#endif
		//printf("%lf\n", Res);
		if (Res > Max)
		{
			Max = Res;
			Index = i;
		}	
	}

	net->SetInput(Inputs[Index], FCDiffTime);
	net->GetOutput(FCDiffTime);
	printf("\nPull je: %d\n", FCDiffTime);

	return Index;
}

void PlayerSkills::FCDribblePerformActions(double *Input)
{
	// dekoduje Input a vykona elementarne akcie..
	
	printf("Time: %d, Uhol: %f ", world->state->GetSimTime(), world->me->GetAbsDir().GetDeg()); 
	// Kopnem do lopty ak {0, 0, 1}
	if (Input[DRIBBLE_KICK] > 0.5)
	{
		// kopeme do lopty
		m_pCommandGen->Kick((float)Input[DRIBBLE_KICK_POWER], (float)Input[DRIBBLE_KICK_ANGLE]);
		printf("Kopem s uhol %lf a silou %lf\n", Input[DRIBBLE_KICK_ANGLE], Input[DRIBBLE_KICK_POWER]);
	} else if (Input[DRIBBLE_TURN] > 0.5)
	{
		// otacame sa
		m_pCommandGen->Turn((float)Input[DRIBBLE_TURN_ANGLE]);
		printf("Otacam sa o uhol %lf\n", Input[DRIBBLE_TURN_ANGLE]);
	} else if (Input[DRIBBLE_DASH] > 0.5)
	{
		// bezime
		m_pCommandGen->Dash((float)Input[DRIBBLE_DASH_POWER]);
		printf("Bezim silou %lf\n", Input[DRIBBLE_DASH_POWER]);
	}
}

void PlayerSkills::FCFixDribbleWeights(const Point& Target)
{
	
	// upravi vahy siete v zavislosti od aktualnej polohy hraca vzhladom na ciel a cas, za ktory to zvladol..
	
	// pociatocne vektory objektov vzhladom k cielu
	Vector	SourceBallPos = FCBSource - FCTarget, 
			SourceMePos = FCPSource - FCTarget;

	// koncove vektory objektov vzhladom k cielu
	Vector	TargetBallPos = world->ball->pos - FCTarget, 
			TargetMePos = world->me->pos - FCTarget;

	
	double dP, dB, dT;

	//dP = (double)SourceMePos.GetLength() / (double)TargetMePos.GetLength();
	//dP = (dP > 1.0)?1.0:dP;

	dP = 1-(((double)SourceMePos.GetLength()-TargetMePos.GetLength()) / (double)SourceMePos.GetLength());
	dP = (dP > 1.0)?1.0:dP;
	
	//dB = (double)SourceBallPos.GetLength() / (double)TargetBallPos.GetLength();
	//dB = (dB > 1.0)?1.0:dB;

	dB = 1-(((double)SourceBallPos.GetLength()-TargetBallPos.GetLength()) / (double)SourceBallPos.GetLength());
	dB = (dB > 1.0)?1.0:dB;


	int currTime = world->state->GetSimTime();
	if (FCDurTime != -1)
	{
		dT = (double)currTime / (double) FCDiffTime;
	} else
	{
		dT = (double)FCDurTime / (double) FCDiffTime;
	}

	dT = (dT > 1.0)?1.0:dT;

	double Demand = (1 - dP)*(1 - dB);//*(1 - dT);
	
	// ulozim nove vahy
	
	//m_pFCDribbleNet->BackProp(Demand, currTime - FCTime);
	m_pFCDribbleNet->BackProp(Demand, FCDiffTime);
	printf("Ohodnotenie %lf\n", Demand);

	// m_pFCDribbleNet->SaveNet(file....);
}

void PlayerSkills::FCGetDribbleInputs(int &N, double **&Input)
{

	int i, j, Ang = 15;
	N = 100;
	double	AngleStep = 3.0, PowerStep = 10.0, MinAngle, MaxAngle;

	double	MinPower = 10.0, //world->params->server->minpower, 
			MaxPower = world->params->server->maxpower/2;

	// uhol ktory zoviera hrac(s jeho aktualnym natocenim) s miestom, kde ma dodribovat
	Angle Pdir = CalcDir(FCTarget - world->me->pos, world->me->GetAbsDir());
	Angle Bdir = CalcDir(world->ball->pos - world->me->pos, world->me->GetAbsDir());

	if (m_bKickable)
	{
		MinAngle = (Pdir.GetDeg() > 0)?(Pdir.GetDeg() - Ang):(Ang - Pdir.GetDeg())*(-1);
		MaxAngle = MinAngle + 2*Ang;
	} else
	{
		MinAngle = (Bdir.GetDeg() > 0)?(Bdir.GetDeg() - Ang):(Ang - Bdir.GetDeg())*(-1);
		MaxAngle = MinAngle + 2*Ang;
	}

	int		Angles = (int)((MaxAngle - MinAngle) / AngleStep),
			Powers = (int)((MaxPower - MinPower) / PowerStep);

	//AngleStep = (MaxAngle < MinAngle)?AngleStep*(-1):AngleStep;
	
	//Angles = (int)((Pdir.GetDeg() - 45) + (Pdir.GetDeg() - 45)) / AngleStep;
	//MinAngle = Pdir.GetDeg() - 45;

	printf("Uhol <%f, %f> = %f \n", MaxAngle, MinAngle, MaxAngle - MinAngle);

	if (m_bKickable)
		N = Powers + Powers + Angles;
	else
		N = Powers + Angles;

	if (Input == NULL)
		Input = new double*[N];
	for (i = 0; i < N; i++)
	{
		Input[i] = new double[DRIBBLE_NET_INPUT];
		memset((void*)Input[i], 0, DRIBBLE_NET_INPUT*sizeof(double));
	}


	i = 0;
	// naplnanie kick-mi
	if (m_bKickable)
	{
		for (j = 0; j < Powers; j++)
			Input[i][DRIBBLE_KICK] = 1.0;
			Input[i][DRIBBLE_KICK_ANGLE] = 0;
			//Input[i][DRIBBLE_KICK_POWER] = MinPower + j*PowerStep;
			Input[i][DRIBBLE_KICK_POWER] = MinPower + ((double)rand()/RAND_MAX)*(MaxPower-MinPower);
			i++;
	}

	// a naplnennie Dash-mi
	for (j = 0; j < Powers; j++)
	{
		Input[i][DRIBBLE_DASH] = 1.0;
		Input[i][DRIBBLE_DASH_POWER] = MinPower + ((double)rand()/RAND_MAX)*(MaxPower-MinPower);
		i++;
	}

	// naplnenie turn-mi
	for (j = 0; j < Angles; j++)
	{
		Input[i][DRIBBLE_TURN] = 1.0;
		Input[i][DRIBBLE_TURN_ANGLE] = MinAngle + ((double)rand()/RAND_MAX)*(MaxAngle - MinAngle);
		i++;
	}

	Vector	BallToPlayer = world->ball->pos - world->me->pos,
			BallToTarget = world->ball->pos - FCTarget,
			TargetToPlayer = FCTarget - world->me->pos;
	
	double	BallX = BallToPlayer.GetX(), BallY = BallToPlayer.GetY(),
			TargetX = TargetToPlayer.GetX(), TargetY = TargetToPlayer.GetY(),
			dB = BallToTarget.GetLength(),
			dP = TargetToPlayer.GetLength(),
			time = TRAINING_DURATION - (world->state->GetSimTime() - FCTime);

	for (i = 0; i < N; i++)
	{
		Input[i][0] = BallToPlayer.GetLength();// BallX;
		Input[i][1] = Bdir.GetDeg();// BallY;
		Input[i][2] = TargetToPlayer.GetLength();// TargetX
		Input[i][3] = Bdir.GetDeg();// TargetY;
		Input[i][DRIBBLE_NET_INPUT - 3] = dB;
		Input[i][DRIBBLE_NET_INPUT - 2] = dP;
		Input[i][DRIBBLE_NET_INPUT - 1] = time;
	}

}

void PlayerSkills::FCDribbleTo(const Point& Target)
{

	int N = 0, Index, i;
	double	**Input = NULL;
	
	// vygenerovanie vstupov
	FCGetDribbleInputs(N, Input);

	// zvolenie toho najlepsieho
	Index = FCEvaluateInputs(m_pFCDribbleNet, N, Input);
		
	// a vykonanie vybratej akcie
	FCDribblePerformActions(Input[Index]);

	for (i = 0; i < N; i++)
		delete [] Input[i];

	delete [] Input;
	
}

void PlayerSkills::FCSetViewWidth(ViewWidth iWidth)
{
	m_pCommandGen->ChangeView(iWidth);
}

#ifdef TRAINING

	void PlayerSkills::OnChangeMode(int time, PlayMode mode)
	{
		PlayerKernel::OnChangeMode(time, mode);


	#ifdef DRIBBLING_TRAINING
		
		if (mode == PM_PlayOn)
		{
			// zaciatok trenovania - a nastavenie hraca do stredu
			FCDiffTime = 0;
			
			FCPSource.SetX(world->me->pos.GetX());
			FCPSource.SetY(world->me->pos.GetY());

			FCBSource.SetX(world->ball->pos.GetX());
			FCBSource.SetY(world->ball->pos.GetY());
		
			FCTarget.SetX(0.0);
			FCTarget.SetY(0.0);
			

			FCTime	=	time + 1;
			FCDurTime = -1;
			FCWriteTime++;
		} else if (mode == PM_TimeOver)
		{
			FCFixDribbleWeights(FCTarget);
			if (FCWriteTime == 30)
			{
				printf("Ukladam!\n");
				m_pFCDribbleNet->SaveAs("vahy.txt");
				FCWriteTime = 0;
				printf("Ulozenuo!\n");
			}
			FCDiffTime = 0;

		}
		
		printf("Uhol je %0.2f\n", world->me->GetAbsDir().GetAbsDeg());
	#endif
	
	}

#endif

bool PlayerSkills::FCBelongsToArea(Point A, Point B, Point E)
{

	double dist = 2;

	double AB = sqrt(pow((A.GetX()-B.GetX()), 2) + pow((A.GetY()-B.GetY()), 2));	// vzdialenost medzi A a B
	double AE = sqrt(pow((A.GetX()-E.GetX()), 2) + pow((A.GetY()-E.GetY()), 2));	// vzdialenost medzi A a E
	double BE = sqrt(pow((B.GetX()-E.GetX()), 2) + pow((B.GetY()-E.GetY()), 2));	// vzdialenost medzi A a E
	double AS = sqrt(pow(((A.GetX()-B.GetX())/2), 2) + pow(((A.GetY()-B.GetY())/2), 2));	// vzdialenost medzi A a B
	double SE = sqrt(pow(((A.GetX()-B.GetX())/2-E.GetX()), 2) + pow(((A.GetY()-B.GetY())/2-E.GetY()), 2));	// vzdialenost medzi A a E
	double AEp = ((E.GetX()-A.GetX())*(B.GetX()-A.GetX()) + (E.GetY()-A.GetY())*(B.GetY()-A.GetY()))/AB;	// vzdialenost medzi A a priemetom E do AB
	double pE = sqrt(AE*AE - AEp*AEp);
		
	return 	(SE <= AS) && (pE < dist) || (AE < dist) || (BE < dist);
}

void	PlayerSkills::FCGetDistances(Point A, Point B, Point E, double* result)
{
	double AB = sqrt(pow((A.GetX()-B.GetX()), 2) + pow((A.GetY()-B.GetY()), 2));	// vzdialenost medzi A a B
	double AE = sqrt(pow((A.GetX()-E.GetX()), 2) + pow((A.GetY()-E.GetY()), 2));	// vzdialenost medzi A a E
	double AEp = ((E.GetX()-A.GetX())*(B.GetX()-A.GetX()) + (E.GetY()-A.GetY())*(B.GetY()-A.GetY()))/AB; // vzdialenost medzi A a priemetom E do AB
	double pE = sqrt(AE*AE - AEp*AEp);

	result[0] = fabs(AB);
	result[1] = fabs(AEp);
	result[2] = fabs(pE);
}

// NASLEDNE FUNKCIE NIE SU PLNE FUNKCNE, JE VHODNE ICH PREPRACOVAT

// Zisti sa, ci moze hrac zachytit loptu. Ak moze, zisti sa kam ma bezat.
bool PlayerSkills::CanInterceptBall(Point* point)
{
	CATCH_DBGN("---> enter CanInterceptBall at %i", world->state->GetSimTime());

	if(world->state->GetPlayMode() != PM_PlayOn)
	{
		if(ShouldKickIn())
			return true;
		return false;
	}

	// Simulate ball path
	List<CatchSimData> data;
	CatchSimBallPath(&data);

	// Evaluate the results of simulation
	bool retval = CatchEvaluateData(point, &data);

	CATCH_DBGN("<--- exit CanInterceptBall");
	
	return retval;
}

// Pomocna funkcia pre CanInterceptBall -> vyhodnocuje simulacne udaje
bool PlayerSkills::CatchEvaluateData(Point* point, List<CatchSimData>* input)
{
	// TODO: different kickable areas for different heteroplayers
	const float karea = 0;//world->params->server->player_size + world->params->server->ball_size + world->params->server->kickable_margin;

	// Zahravame my? Ak ano, ignoruj superov, pretoze nemozu kopnut do lopty.
	bool ourKickIn = false;
	if(world->state->IsOurKickIn())
		ourKickIn = true;
//	CATCH_DBGN("  ourKickIn = %s", ourKickIn ? "true" : "false");
	
	int friendCaught = 0, enemyCaught = 0, meCaught = 0;
	float friendBalance = 0.f, enemyBalance = 0.f, meBalance = 0.f;

	int catchTime = INT_MAX;
	int currTime = 0;
	int i;

	// Ktori hraci su uz blizko pri lopte?
	bool* friendGotBall = (bool*) malloc(input->GetFirst()->data->GetFriendCnt() * sizeof(bool));
	
	for(i = 0; i < input->GetFirst()->data->GetFriendCnt(); i++)
		friendGotBall[i] = false;
	
	bool* enemyGotBall = (bool*) malloc(input->GetFirst()->data->GetEnemyCnt() * sizeof(bool));

	for(i = 0; i < input->GetFirst()->data->GetEnemyCnt(); i++)
		enemyGotBall[i] = false;

	// Spracuju sa simulacne udaje
	LIST_FOR_EACH(CatchSimData, cycle, *input)
	{
		for(i = 0; i < cycle->data->GetFriendCnt(); i++)
		{
			if(!friendGotBall[i] && cycle->data->GetFriendDif(i) <= karea)
			{
				friendCaught++; friendBalance++;
				friendGotBall[i] = true;
				CATCH_DBGN("  friend %i caught at %i: caught %i balance %3.2f", world->friends[i].GetUniform(), currTime, friendCaught, friendBalance);
				
				if(catchTime == INT_MAX)
				{
					catchTime = currTime;
					CATCH_DBGN("  catchTime = %i", catchTime);
				}
			}
		}

		if(!ourKickIn) for(i = 0; i < cycle->data->GetEnemyCnt(); i++)
		{
			if(!enemyGotBall[i] && cycle->data->GetEnemyDif(i) <= karea)
			{
				enemyCaught++; enemyBalance++;
				enemyGotBall[i] = true;
				CATCH_DBGN("  enemy %i caught at %i: caught %i balance %3.2f", world->enemies[i].GetUniform(), currTime, enemyCaught, enemyBalance);

				if(catchTime == INT_MAX)
				{
					catchTime = currTime;
					CATCH_DBGN("  catchTime = %i", catchTime);
				}
			}
		}

		if(cycle->data->GetMyDif() <= karea)
		{
			*point = cycle->data->ballPos;
			meCaught++; meBalance++;
			CATCH_DBGN("  me caught at %i, BREAK", currTime);
			
			if(catchTime == INT_MAX)
			{
				catchTime = currTime;
				CATCH_DBGN("  catchTime = %i", catchTime);
			}

			break;
		}

		// Vela casu uplynulo, kym prvy hrac chytil loptu
		if(currTime - catchTime >= AFTER_CATCH_CYCLES)
		{
			//if(meCaught == 0) *point = cycle->data->ballPos;
			CATCH_DBGN("  currTime = %i, catchTime = %i, BREAK", currTime, catchTime);
			break;
		}
		
		// Prilis vela hracov chytilo loptu
//		if(friendCaught > GetCoachAdjParams()->GetFriendMaxCatchers()) break;
//		if(enemyCaught > GetCoachAdjParams()->GetEnemyMaxCatchers()) break;

		friendBalance *= world->variables->GetFriendBallCtrlBonus();
		enemyBalance *= world->variables->GetEnemyBallCtrlBonus();
		meBalance *= world->variables->GetFriendBallCtrlBonus();
		
		currTime++;
	}

	CATCH_DBGN("  SUMMARY:");
	CATCH_DBGN("  meCaught: %i, friendCaught: %i, enemyCaught: %i", meCaught, friendCaught, enemyCaught);
	CATCH_DBGN("  meBalance: %3.2f, friendBalance: %3.2f, enemyBalance: %3.2f", meBalance, friendBalance, enemyBalance);

	int dbgt = 0;
	LIST_FOR_EACH(CatchSimData, cycle2, *input)
	{
		dbgt++;
		if(cycle2->data->GetMyDif() <= karea)
		{
			CATCH_DBGN("  just DBG: I would catch at %i", currTime);
			break;
		}
	}

	free(friendGotBall); free(enemyGotBall);

	if(meCaught == 0)
		return false;

	// Urcite dokazem zachytit loptu, mam?
	// Kto bude kontrolovat loptu, ak ju budem ignorovat?
	float balance = friendBalance - enemyBalance;
	CATCH_DBGN("  balance: %3.2f", balance);

	if(!ourKickIn)
	{
		if(balance + meBalance < -world->variables->GetTooMuchEnemyCnt())
		{
			// Prilis vela superov. 
			CATCH_DBGN("  RESULT: Too many enemies");
			CATCH_DBGN("  IGNORE");
			return false;
		}
		else if(balance < world->variables->GetTooMuchFriendCnt())
		{
			// Je to 50 na 50
			CATCH_DBGN("  RESULT: Fifty-fifty");
		}
		else
		{
			// Dost spoluhracov. Netreba viac pomahat
			CATCH_DBGN("  RESULT: Too many friends");
			CATCH_DBGN("  IGNORE");
			return false;
		}
	}

	// Ak sme sa dostali az sem, chytam loptu ja
	CATCH_DBGN("  CATCHING at (%3.2f, %3.2f)", point->GetX(), point->GetY());
	return true;
}

// Pomocna funkcia pre CanInterceptBall -> simuluje drahu lopty
void PlayerSkills::CatchSimBallPath(List<CatchSimData>* output)
{
	int i;
	Ball tmpBall(*world->ball);

	bool ourKickIn = false;
	if(world->state->IsOurKickIn()) ourKickIn = true;

	// Predpokladane rychlosti hracov
	const float myDashPower = RecommDashPower(true) * world->me->GetEffort();
	const float friendDashPower = 
		world->variables->GetFriendDashAdjust() * (world->params->server->stamina_inc_max + world->params->server->maxpower) / 2;
	const float enemyDashPower = 
		world->variables->GetEnemyDashAdjust() * world->params->server->maxpower;

	// Pocty znamych spoluhracov aj superov
	int friendCnt = world->friends.GetSize();
	int enemyCnt = world->enemies.GetSize();

	// Vzdialenosti bezane vsetkymi hracmi
	float myDist = 0;
	
	float* friendDists = (float*) malloc(sizeof(float) * friendCnt);
	
	for(i = 0; i < friendCnt; i++)
		friendDists[i] = 0;
	
	float* enemyDists = (float*) malloc(sizeof(float) * enemyCnt);
	
	for(i = 0; i < enemyCnt; i++)
		enemyDists[i] = 0;
	
	// Sucastne rychlosti vsetkych hracov
	float mySpd = world->me->vel.GetLength();
	float* friendSpds = (float*) malloc(sizeof(float) * friendCnt);
	
	for(i = 0; i < friendCnt; i++)
		friendSpds[i] = world->friends[i].vel.GetLength();
	
	float* enemySpds = (float*) malloc(sizeof(float) * enemyCnt);
	
	for(i = 0; i < enemyCnt; i++)
		enemySpds[i] = world->enemies[i].vel.GetLength();

	// Pre kazdy cyklus a kazdeho hraca. Uloz jeho vzdialenost od lopty. Pre kazdy cyklus uloz poziciu lopty.
	// Dopredu nevieme potrebny pocet cyklov.
	// Simuluje sa draha lopty. Ukonci sa to, ked dobehneme loptu alebo, ked opusti ihrisko. Ked chytame loptu,
	// simulujeme este niekolko krokov, aby sme uvideli, ci niekto iny ju nedobehne v tom istom cykle.

	int caught = -1;
	int cycle = 0;

	float hack_firstCycleDistNoTurn = 0.f;
	float hack_firstCycleDistWithTurn = 0.f;

	while(true)
	{
		cycle++;
		tmpBall.Step();

		// Uloz drahu lopty v tomto cykle
		CatchSimData* data = new CatchSimData(friendCnt, enemyCnt);
		data->ballPos = tmpBall.pos;

		// Pre kazdeho hraca, vypocitaj jeho vzdialenost, ktoru moze odbehnut v tomto cykle. Kompenzuj otacanie
		// v prvom cykle, ak je to potrebne. Potom vypocitaj vzdialenost hraca od lopty.

		// Tento hrac
		if(cycle == 0)
		{
			// Vypocitaj prvy cyklus s a bez otacania
			CatchSimRunningPlayer(myDashPower, &mySpd, &hack_firstCycleDistNoTurn);
			CatchSimRunningPlayer(0, &mySpd, &hack_firstCycleDistWithTurn);
		}
		else
		{
			CatchSimRunningPlayer(myDashPower, &mySpd, &myDist);
		}

		// Kde budem, ked sa neotocim v prvom cykle?
		Point noTurnPos = world->me->pos + PolarVector(hack_firstCycleDistNoTurn + myDist, world->me->GetAbsDir());

		float ballDiffNoTurn = (tmpBall.pos - noTurnPos).GetLength();
		if(DoLinesIntersect(world->me->pos, noTurnPos - world->me->pos, tmpBall.pos, tmpBall.vel))
			ballDiffNoTurn = -ballDiffNoTurn;
		float ballDiffWithTurn = (tmpBall.pos - world->me->pos).GetLength() - myDist - hack_firstCycleDistWithTurn;

#define SOME_DELTA world->params->server->player_size + world->params->server->ball_size + world->params->server->kickable_margin
		bool dbgWhichOne;
		if(ballDiffNoTurn < SOME_DELTA)
		{
			// Nemusime sa otocit
			dbgWhichOne = true;
			data->ballPos = noTurnPos;
			data->SetMyDif(ballDiffNoTurn);
		}
		else
		{
			// Musime sa otocit
			dbgWhichOne = false;
			data->SetMyDif(ballDiffWithTurn);
		}

		// Skontrolujme, ci som chytil loptu
		if(data->GetMyDif() <= 0 && caught == -1)
		{
			CATCH_DBGN("  SIMULATION: I caught the ball (%3.2f, %3.2f) after %i", data->ballPos.GetX(), data->ballPos.GetY(), cycle);
			CATCH_DBGN("  without turn: I am at (%3.2f, %3.2f), diff is %3.2f", noTurnPos.GetX(), noTurnPos.GetY(), ballDiffNoTurn);
			CATCH_DBGN("  with turn: diff is %3.2f", ballDiffWithTurn);
			CATCH_DBGN("  using: %s, cond %s: %3.2f < %3.2f", dbgWhichOne ? "no turn" : "with turn", dbgWhichOne ? "TRUE" : "FALSE", ballDiffNoTurn, SOME_DELTA);
			caught = 0;
		}


		// Spoluhraci
		//if(friendCnt > 10) DBGN("friends %i", friendCnt);
		for(i = 0; i < friendCnt; i++)
		{
			CatchSimRunningPlayer(friendDashPower, &friendSpds[i], &friendDists[i]);
			data->SetFriendDif(
				i, (tmpBall.pos - world->friends[i].pos).GetLength() - friendDists[i]);
		}


		// Superovia
		//if(count > 11) DBGN("enemies %i", count);
		for(i = 0; i < enemyCnt; i++)
		{
			CatchSimRunningPlayer(enemyDashPower, &enemySpds[i], &enemyDists[i]);
			data->SetEnemyDif(
				i, (tmpBall.pos - world->enemies[i].pos).GetLength() - enemyDists[i]);
		}

		output->Add(data);

	
		// Ak je lopta mimo, ignoruj ju. Ale iba ak by to nebol jeden z nasich zahravacich modov.
		if(!ourKickIn &&
			((tmpBall.pos.GetX() < world->params->flags[LT].GetX() - POS_ERROR_DELTA) ||
			(tmpBall.pos.GetY() < world->params->flags[LT].GetY() - POS_ERROR_DELTA) ||
			(tmpBall.pos.GetX() > world->params->flags[RB].GetX() + POS_ERROR_DELTA) ||
			(tmpBall.pos.GetY() > world->params->flags[RB].GetY() + POS_ERROR_DELTA)))
		{
//			CATCH_DBGN("  ignoring ball - out");
			break;
		}
		
		// Uz sme chytlili loptu a simulovali niekolko AFTER_CATCH_CYCLES navyse?
		if(caught >= AFTER_CATCH_CYCLES)
			break;
		
		// Ak uz mame loptu, zvys pocitadlo. Ak nemame loptu, pocitadlo je -1. Ked ju chytime, nastavime pocitadlo na 0.
		if(caught >= 0)
			caught++;
	}

	free(friendDists); free(enemyDists);
	free(friendSpds); free(enemySpds);
}

// Simuluje hraca pocas jedneho cyklu. Vstupna rychlost hraca a vzdialenost sa pouziju teraz a koeficient sprintu
// v dalsom cykle. Vystup je nova rychlost a kumulovana vzdialenost. Vyuziva sa pre chytanie lopty.
void PlayerSkills::CatchSimRunningPlayer(float dashPower, float* speed, float* distance)
{
	float acceleration = dashPower * world->params->server->dash_power_rate;
	
	if(acceleration > world->params->server->player_accel_max)
		acceleration = world->params->server->player_accel_max;

	*speed += acceleration;
	
	if(*speed > world->params->server->player_speed_max)
		*speed = world->params->server->player_speed_max;

	*distance += *speed;
	*speed *= world->params->server->player_decay;
}


