// PlayerTactics.h

// Definovanie funkcii bezprostredne suvisiacich s taktickou strankou hraca (napr. spravanie sa hraca v roznych situaciach)

#ifndef PLAYERTACTICS_H
#define PLAYERTACTICS_H

#include "Player/PlayerSkills.h"
//#include "Common/Main.h"
#include "Formation.h"
#include "Player/Interfaces/Tactics.h"
#include <stdio.h>

//#define KICKIN_DBGN DBGN
#define KICKIN_DBGN DO_NOTHING

enum LPT_FormationType
{
	LPT_FR_LJON = 1,
  LPT_FR_1,
  LPT_FR_2,
  LPT_FR_3,
  LPT_FR_4,
  LPT_FR_5,
  LPT_FR_6,
  LPT_FR_7,
  LPT_FR_8,
  LPT_FR_9,
  LPT_FR_10,
  LPT_FR_11,
  LPT_FR_12,
	LPT_FR_QCRN,
  LPT_FR_FIRST = LPT_FR_LJON,
  LPT_FR_LAST = LPT_FR_QCRN
};

// Modul taktickej stranky hraca
class PlayerTactics : public PlayerSkills, public TacticsInterface 
{
//==========================
//Loptosi ============
	bool LPT_AttentionToSet;
	bool LPT_AmILeader;
	int LPT_LeaderUniform;
	int LPT_LastSendTime;
	int LPT_LastChangeTime;
	LPT_FormationType LPT_CurrentFormation;
	//FILE* LPT_LeaderFileOut;
//Loptosi ============
//==========================

	Formation m_cFormation;

	// Ignoruj loptu zatial co brankar vykopava
	bool letGoalieRest;

	// Uhol scanovania pred vykopavanim
	float scanBeforeKickOffAngle;

	
//----------------->
	// Kolko cyklov uz trvaju jednotlive mody?
	int nCornerKickModeDuration;
	int nFreeKickModeDuration;
	int nIndFreeKickModeDuration;
	int nKickInModeDuration;

	// Posledny cyklus, kedy sa pocital optimalny driblovaci bod
	int nLastDribblePointCalculationCycle;

	// Bod pre driblovanie, ktory sa pouziva nielen pre urcenie optimalneho bodu pre dribling
	double lfDribbleX, lfDribbleY;
//<-----------------

	// Vrati formaciu
    int GetFormation(const char* formation);
   
	//
	int GetUnit(int PlayerCount, char type);

protected:
	// FUNKCIE SPRAVANIA SA
	
	// Spravanie hraca s loptou
	virtual void BehaveWithBall();
	virtual void BehaveWithoutBall();
	virtual void BehaveWithBallGoalie();
	virtual void BehaveWithoutBallGoalie();

	// Vypocita a vrati vzdialenost k branke
	float DistanceToGoal(const Point&) const;

	// Hrac preskuma najblizsie okolie okolo seba otocenim sa okolo vlastnej osi 
	void ExamineSurroundings();

	// Hrac odbehne z ofsajdu
	void ExitOffside();

	// Vrati NULL, ak superov brankar nie je v hracovom svete
	Player* GetEnemyGoalie() const;

	// Vrati domovsku poziciu
	virtual Point GetHomePos();
	
	//Zisti silu potrebnu na kopnutie lopty na urcite miesto za urcity pocet cyklov
	float GetKickPower(const Point& ciel, int steps);
	
	// Vrati vzdialenost k nablizsiemu superovemu hracovi
	float GetNearestEnemyDistance(const Point& target);

	// Vrati najblizsi index supera
	Player GetNearestEnemy(const Point& target);

	// Vrati najblizsi index spoluhraca
	int GetNearestFriend(const Point& target);
	
	// Vrati vzdialenost k najblizsiemu spoluhracovi
	float GetNearestFriendDistance(const Point& target);
	
	// Vrati hracovu vzdialenost k lopte
//	float GetPlayerBallDistance(int who);
	
	// Chod na najlepsie miesto, kde ti niekto moze prihrat loptu
	bool GotoPositionToPass();

	// Zmeni formaciu
    void ChangeFormation(const char* formation);

	// Brani super v nasej prihravke na urcite miesto?
	bool IsBlockingEnemy(const Point& target);
	
	// Je urcene miesto ofsajd?
	bool IsOffsidePosition(Point target);

	// Tato metoda riadi vypocitavanie optimalneho driblovacieho bodu
	bool ManageDribblePoint();
	
	// Su nejaki superi okolo? Ak nie, tak by mal hrac driblovat.
	bool NoEnemiesAround() const;

	// Zmeni handler hracieho modu
	virtual void OnChangeMode(int time, PlayMode mode);

	// Hrac si predkopne loptu popri superovi
	void OutPlayOpponent();

	// Prihranie beziacemu hracovi
	bool PassIntoRun();
	
	// Beh na takticku poziciu alebo sledovanie lopty, ak uz tam hrac je
	void Patrol();
	
	// Prihra na dane miesto kratkym kopnutim, ak sa tam dostane skor ako superi
	bool ShortKick(const Point&);
	
	// Vrati true alebo false, podla toho, ci by tento hrac mal zahravat
	virtual bool ShouldKickIn();

	// Mal by hrac vystrelit na branku?
	bool ShouldKickOnGoal();
	
	// Vrati true a hrac dribluje k bodu, ak super nemoze kopnut do lopty. Inak vrati false.
	bool SmartDribleTo(const Point&);

	// Rozohranie lopty spoluhracov zo zahravacich modov
	bool StartPlayOnByPassToOurPlayer();
	
	// Hrac sa bude snazit neprijat nahravku
	bool UnblockMyself(Vector* shift);
			
public:

	// Konstruktor
	PlayerTactics();

	// Destruktor
	virtual ~PlayerTactics();

	// Vrati tento takticky interfejs
	virtual Interface* GetInterface(int task) const;
	// Inicializovanie taktickeho modulu
	void OnInit(RealSide realSide, unsigned uniform, PlayMode iMode);
	
//======================================
	//Loptosi ============
	void OnHear(int time, float direction, bool ourTeam, int uniform, const char* msg);
	void RunPlayOn();
	void RunPlayOff();

	// Metody
	void LPT_SendZmenaFormacieLeaderMSG();              // LFC
	void LPT_SendZmenaFormaciePlayerMSG();              // PFC
	void LPT_SendTySiLeaderMSG(int uniformNumber);      // LTY
	void LPT_SendOkSomLeaderMSG();                      // LOK
	void LPT_SendLeaderMSG();                           // LJE
	void LPT_processLeaderMSG(int time, int uniform, const char* msg);
	bool LPT_GetAmILeader();
	int LPT_GetLeaderUniform();
//Loptosi ============
//===================================

	// Vrati ofsajdovu ciaru
	float GetOffsideLine() const;

	// Vrati hraca, ktory je najvhodnejsie na nahravanie. -1 => ziadny hrac
	int GetPassAbleFriend();

	// Je hrac v ofsajde?
	bool InOffsidePos();
	
	/// \return is player responsible for the ball?
	//bool DependentForBall() const;


// ====================== FC Farmari ======================
private:

#ifdef PASS_TRAINING
	char	FFileName[20];
	FILE	*ff;
	bool	chytilsom;
#endif

public:
	FCNetwork	*m_PassNet;

	int FCGetPassAbleFriend();
	void RealToBin(double num, double* buff);
};

#endif // PLAYERTACTICS_H

